import collections, abc
import functools
import numpy as np
import itertools

from .. import Devutils as dev

# from scipy.optimize.linesearch import line_search_armijo
from . import VectorOps as vec_ops
from . import Misc as misc
from . import TransformationMatrices as tfs
from . import SetOps as set_ops

__all__ = [
    "iterative_step_minimize",
    "iterative_chain_minimize",
    "GradientDescentStepFinder",
    "NewtonStepFinder",
    "QuasiNewtonStepFinder",
    "ConjugateGradientStepFinder",
    "EigenvalueFollowingStepFinder",
    "NudgedElasticBandStepFinder",
    "jacobi_maximize",
    "LineSearchRotationGenerator",
    "GradientDescentRotationGenerator",
    "OperatorMatrixRotationGenerator",
    "displacement_localizing_rotation_generator"
]
default_jacobian_step_finder = 'conjugate-gradient'
default_hessian_step_finder = 'newton'
def lookup_method_name(method):
    return {
        'conjugate-gradient':ConjugateGradientStepFinder,
        'newton':NewtonStepFinder,
        'quasi-newton':QuasiNewtonStepFinder,
        'gradient-descent':GradientDescentStepFinder,
        'neb':NudgedElasticBandStepFinder
    }.get(method)
def get_step_finder(spec,
                    method=None,
                    jacobian=None,
                    hessian=None,
                    **extra_init):
    if not (isinstance(spec, dict) or hasattr(spec, 'supports_hessian')) and (
            jacobian is not None
            or method is not None
    ):
        spec = {
            'method':method,
            'func':spec,
            'jacobian':jacobian,
            'hessian':hessian
        }
    if isinstance(spec, dict):
        spec = spec.copy() # don't mutate user data
        method = spec.pop('method', method)
        func = spec.pop('func')
        if jacobian is None:
            jacobian = spec.pop('jacobian')
        else:
            jacobian = spec.pop('jacobian', jacobian)
        hessian = spec.pop('hessian', hessian)
        if method is None:
            if hessian is not None:
                method = default_hessian_step_finder
            else:
                method = default_jacobian_step_finder
        if isinstance(method, str):
            test_method = lookup_method_name(method)
            if test_method is None:
                raise ValueError(f"can't determine appropriate step finder for '{method}")
            method = test_method
        if hessian is not None and method.supports_hessian:
            spec = method(func, jacobian, hessian, **dict(spec, **extra_init))
        else:
            spec = method(func, jacobian, **dict(spec, **extra_init))
    return spec
oscillation_overlap_cutoff_min = 0.95
oscillation_overlap_cutoff_max = 1.05
def iterative_step_minimize_step(step_predictor,
                                 guess, mask, tol,
                                 orthogonal_projector, orthogonal_projection_generator,
                                 region_constraints, unitary, max_displacement, max_displacement_norm,
                                 generate_rotation, prev_steps, max_gradient_error, termination_function
                                 ):
    if unitary:
        projector = vec_ops.orthogonal_projection_matrix(guess.T)
        if orthogonal_projector is not None:
            projector = projector @ orthogonal_projector[np.newaxis]
        if orthogonal_projection_generator is not None:
            projector = projector @ orthogonal_projection_generator(guess)
    elif orthogonal_projector is not None:
        projector = orthogonal_projector[np.newaxis]
        if orthogonal_projection_generator is not None:
            projector = projector @ orthogonal_projection_generator(guess)
    elif orthogonal_projection_generator is not None:
        projector = orthogonal_projection_generator(guess)
    else:
        projector = None
    step, grad = step_predictor(guess, mask, projector=projector)
    if projector is not None:
        # just in case, will usually lead to non-convergence
        step = (step[..., np.newaxis, :] @ projector).reshape(step.shape)
        grad = (grad[..., np.newaxis, :] @ projector).reshape(step.shape)

    if max_displacement is not None:
        step_sizes = np.max(np.abs(step), axis=1)
        step *= max_displacement / np.clip(step_sizes, max_displacement, None)
    if max_displacement_norm is not None:
        step_sizes = np.linalg.norm(step, axis=1)
        step *= max_displacement / np.clip(step_sizes, max_displacement, None)
    if region_constraints is not None:
        #TODO: introduce this as a scaling
        max_step = region_constraints[np.newaxis, :, 1] - guess
        min_step = region_constraints[np.newaxis, :, 0] - guess
        step = np.array([
            np.clip(s, smin, smax)
            for s,smin,smax in zip(guess, min_step, max_step)
        ])
    if isinstance(mask, tuple):
        mask = mask[0]

    if max_gradient_error:
        errs = np.max(np.abs(grad), axis=1)
    else:
        errs = np.linalg.norm(grad, axis=1)
    rem = np.arange(len(mask))
    done = np.where(errs < tol)[0]
    if len(done) > 0:  # easy check
        rem = np.delete(rem, done)
        mask = np.delete(mask, done)
        if len(mask) == 0:
            return step, errs, [], done
    if termination_function is not None:
        done = np.where(termination_function(guess[rem,], step[rem,], mask))
        if len(done) > 0:  # easy check
            rem = np.delete(rem, done)
            mask = np.delete(mask, done)
            if len(mask) == 0:
                return step, errs, [], done

    step = step[rem,]

    if prev_steps is not None:
        prev_steps = prev_steps[mask,]
        cur_norms = np.linalg.norm(step, axis=1)[:, np.newaxis]
        pre_norms = np.linalg.norm(prev_steps, axis=2)
        overlaps = vec_ops.vec_tensordot(step, prev_steps, axes=[-1, -1], shared=1) / (cur_norms * pre_norms)
        overlaps = np.reshape(overlaps, prev_steps.shape[:2])
        osc = np.where(
            np.all(
                np.logical_and(
                    oscillation_overlap_cutoff_min < abs(overlaps),
                    abs(overlaps) < oscillation_overlap_cutoff_max
                ),
                axis=1
            )
        )

        if len(osc[0]) > 0:
            dist = np.reshape(overlaps[osc][:, np.newaxis, :] @ pre_norms[osc][:, :, np.newaxis], cur_norms.shape)
            osc2 = np.where(dist < 0)[:1]
            if len(osc2[0]) > 0:
                osc = osc[0][osc2]
                scaling = np.min([cur_norms[osc2], -dist[osc2] / 2], axis=0) / cur_norms[osc2]
                step[osc,] = step[osc,] * scaling

    if unitary:
        step = step
        norms = np.linalg.norm(step, axis=1)
        # norms = norms[rem,]
        v = np.linalg.norm(guess[rem,], axis=-1)
        r = np.sqrt(norms ** 2 + v ** 2)
        step = guess[rem,] * (1/r[:, np.newaxis] - 1) + step / r[:, np.newaxis]
        if generate_rotation:
            raise NotImplementedError(">3D rotations are complicated")
            axis = vec_ops.vec_crosses(g, guess[mask,])[0]
            ang = np.arctan2(norms / r, v / r)
            rot = tfs.rotation_matrix(axis, ang)
            rotations = rot @ rotations
    else:
        step = step[rem,]


    return step, errs, mask, done

def iterative_step_minimize(
        guess,
        step_predictor,
        jacobian=None,
        hessian=None,
        *,
        method=None,
        unitary=False,
        generate_rotation=False,
        dtype='float64',
        orthogonal_directions=None,
        orthogonal_projection_generator=None,
        region_constraints=None,
        function=None,
        max_displacement=None,
        max_displacement_norm=None,
        oscillation_damping_factor=None,
        termination_function=None,
        prevent_oscillations=None,
        tol=1e-8,
        use_max_for_error=True,
        max_iterations=100,
        convergence_metric=None,
        track_best=False,
        logger=None
):
    logger = dev.Logger.lookup(logger)

    step_predictor = get_step_finder(step_predictor,
                                     method=method,
                                     jacobian=jacobian,
                                     hessian=hessian,
                                     logger=logger)
    guess = np.array(guess, dtype=dtype)
    base_shape = guess.shape[:-1]
    guess = guess.reshape(-1, guess.shape[-1])
    if track_best:
        best = guess.copy()
        best_errs = np.full(guess.shape[0], -1, dtype=float)
        if function is not None:
            best_vals = np.full(guess.shape[0], np.inf, dtype=float)
        else:
            best_vals = None
    else:
        best = None
        best_errs = None
        best_vals = None
    its = np.zeros(guess.shape[0], dtype=int)
    errs = np.zeros(guess.shape[0], dtype=float)
    mask = np.arange(guess.shape[0])
    if orthogonal_directions is not None:
        orthogonal_directions = vec_ops.orthogonal_projection_matrix(orthogonal_directions)
    if unitary and generate_rotation:
        rotations = vec_ops.identity_tensors(guess.shape[:-1], guess.shape[-1])
    else:
        rotations = None

    if prevent_oscillations is None:
        prevent_oscillations = (
                unitary
                or orthogonal_projection_generator is not None
                or orthogonal_directions is not None
        )
    if prevent_oscillations is True:
        prevent_oscillations = 1
    prev_step = None
    prev_errs = None

    converged = True
    for i in range(max_iterations):
        with logger.block(tag=f"Iteration {i}"):
            step, step_errs, new_mask, _ = iterative_step_minimize_step(
                step_predictor, guess[mask,],
                mask, tol,
                orthogonal_directions, orthogonal_projection_generator,
                region_constraints, unitary, max_displacement, max_displacement_norm,
                generate_rotation, prev_step, use_max_for_error, termination_function
            )
            if best is not None:
                if i == 0:
                    if best_vals is not None:
                        best_vals[mask,] = function(guess[mask,], mask)
                    best_errs[mask,] = step_errs
                    best[mask,] = guess[mask,]
                else:
                    if best_vals is not None:
                        new_vals = function(guess[mask,], mask)
                        improved = np.where(new_vals < best_vals[mask,])
                        if len(improved[0]) > 0:
                            imp_mask = mask[improved]
                            best_vals[imp_mask,] = new_vals[improved,]
                            best[imp_mask,] = guess[imp_mask,]
                            best_errs[imp_mask,] = step_errs[improved,]
                    else:
                        improved = np.where(step_errs < best_errs[mask,])
                        if len(improved[0]) > 0:
                            imp_mask = mask[improved]
                            best_errs[imp_mask,] = step_errs[improved,]
                            best[imp_mask,] = guess[imp_mask,]


            logger.log_print("Predicted steps: {step}", step=step)
            logger.log_print("Step errors: {errs}", errs=step_errs)
            if prevent_oscillations:
                if prev_step is None:
                    prev_step = np.random.uniform(size=(step.shape[0], prevent_oscillations, step.shape[1])).astype(step.dtype)
                p = i % prevent_oscillations
                prev_step[new_mask, p] = step
            if (not unitary) and oscillation_damping_factor is not None:
                if prev_errs is None:
                    prev_errs = step_errs
                else:
                    prev_errs[new_mask,] = step_errs

                inc_err = np.where(step_errs >= prev_errs[new_mask] - 1e-6) # add some flexibility
                if len(inc_err[0]) > 0:
                    oscillation_damping_factor /= 2
                else:
                    oscillation_damping_factor = np.min([oscillation_damping_factor * 1.2, 1.0])
                step = step * oscillation_damping_factor
                logger.log_print("Oscillation damping factor: {damp}", damp=oscillation_damping_factor)

            errs[mask,] = step_errs
            if len(new_mask) == 0:
                break

            mask = new_mask
            guess[mask,] += step
            its[mask,] += 1
    else:
        converged = False
        its[mask,] = max_iterations
        if best is not None:
            guess[mask,] = best[mask,]
            errs[mask,] = best_errs[mask,]

    guess = guess.reshape(base_shape + (guess.shape[-1],))
    errs = errs.reshape(base_shape)
    its = its.reshape(base_shape)

    if unitary and generate_rotation:
        res = (guess, rotations), converged, (errs, its)
    else:
        res = guess, converged, (errs, its)

    return res

default_chain_step_finder='neb'
def iterative_chain_minimize(
        chain_guesses, step_predictors,
        jacobian=None,
        hessian=None,
        *,
        method=None,
        unitary=False,
        generate_rotation=False,
        dtype='float64',
        orthogonal_directions=None,
        orthogonal_projection_generator=None,
        prevent_oscillations=None,
        region_constraints=None,
        convergence_metric=None,
        termination_function=None,
        reparametrizer=None,
        max_displacement=None,
        max_displacement_norm=None,
        tol=1e-8, max_iterations=100,
        use_max_for_error=True,
        periodic=False,
        reembed=None,
        embedding_options=None,
        fixed_images=None,
        logger=None
):
    guesses = np.array(chain_guesses, dtype=dtype)
    base_shape = guesses.shape[:-2]
    guesses = guesses.reshape((-1,) + guesses.shape[-2:])
    if reembed is None:
        reembed = embedding_options is not None
    if embedding_options is None:
        embedding_options = {}

    try:
        iter(step_predictors)
    except TypeError:
        step_predictors = [step_predictors] * guesses.shape[-2]
    step_predictors = [
        get_step_finder(predictor,
                        method=default_chain_step_finder if method is None else method,
                        jacobian=jacobian,
                        hessian=hessian)
        for predictor in step_predictors
    ]
    its = np.zeros(guesses.shape[0], dtype=int)
    errs = np.zeros(guesses.shape[0], dtype=float)
    submasks = np.full(guesses.shape[:2], True, dtype=bool)
    if fixed_images is not None:
        submasks[..., fixed_images] = False
    mask = np.arange(guesses.shape[0])
    if orthogonal_directions is not None:
        orthogonal_directions = vec_ops.orthogonal_projection_matrix(orthogonal_directions)
    # if unitary and generate_rotation:
    #     rotations = vec_ops.identity_tensors(guess.shape[:-1], guess.shape[-1])
    # else:
    #     rotations = None

    if prevent_oscillations is None:
        prevent_oscillations = (
                unitary
                or orthogonal_projection_generator is not None
                or orthogonal_directions is not None
        )
    if prevent_oscillations:
        prev_step = np.zeros_like(guesses)
    else:
        prev_step = None
    nimg = guesses.shape[-2]
    n = nimg - 1

    if fixed_images is None:
        fixed_images = []

    if reparametrizer is not None:
        image_numbers = np.full(guesses.shape[0], nimg, dtype=int)
    else:
        image_numbers = None

    converged = True
    for i in range(max_iterations):
        # for each unoptimized chain, we only move image `j` if
        # or it's neihbors weren't optimized at the last step, every time an image is moved
        # it's neighbors are marked as moveable again
        errs[mask,] = 0
        for j in range(nimg):
            submask = mask[submasks[mask,][:, j]]
            prev_im = j-1
            next_im = j+1
            if j == 0:
                if periodic:
                    prev_im = n
                else:
                    prev_im = None
            elif j == n:
                if periodic:
                    next_im = 0
                else:
                    next_im = None


            step_predictor = step_predictors[j]
            ps = prev_step[submask,][:, j] if prev_step is not None else prev_step
            step, step_errs, new_mask, done = iterative_step_minimize_step(
                step_predictor, guesses[submask],
                (submask, (j, prev_im, next_im)), tol,
                orthogonal_directions, orthogonal_projection_generator,
                region_constraints, unitary, max_displacement, max_displacement_norm,
                generate_rotation, ps,
                use_max_for_error, termination_function
            )
            if prevent_oscillations:
               prev_step[new_mask, j] = step

            # set which chains are done or not
            done = submask[done]
            submasks[done, j] = False
            if j not in fixed_images:
                submasks[new_mask, j] = True
            if j == 0:
                if j + 1 not in fixed_images:
                    submasks[new_mask, j+1] = True
                if periodic and n not in fixed_images: #TODO: add O(1) check
                    submasks[new_mask, n] = True
            elif j == n:
                if j-1 not in fixed_images:
                    submasks[new_mask, j-1] = True
                if periodic and 0 not in fixed_images:
                    submasks[new_mask, 0] = True
            else:
                if j+1 not in fixed_images:
                    submasks[new_mask, j+1] = True
                if j-1 not in fixed_images:
                    submasks[new_mask, j-1] = True

            errs[mask,] += step_errs
            guesses[submask, j] += step
            if reembed: # implies Cartesian
                from .CoordinateFrames import eckart_embedding

                if j == 0:
                    if periodic:
                        ref = n
                    else:
                        ref = 1
                else:
                    ref = j-1
                ref = guesses[submask, ref]
                coords = guesses[submask, j]
                emb = eckart_embedding(
                    ref.reshape(ref.shape[0], -1, 3),
                    coords.reshape(coords.shape[0], -1, 3),
                    **embedding_options
                )
                guesses[submask, j] = emb.coordinates.reshape(coords.shape[0], -1)

        if reparametrizer is not None:
            # for methods that want to satisfy some density function on
            # the number of images
            new_chains, is_static, fixed_images = reparametrizer(guesses[mask,], fixed_images)
            nimg_new = new_chains.shape[1]
            if nimg_new != nimg:
                guesses = np.pad(guesses,  [[0, 0], [0, nimg_new-nimg], [0, 0]])
                submasks = np.pad(submasks, [[0, 0], [0, nimg_new-nimg]])
                submasks[mask, :] = True
                guesses[mask,] = new_chains
                image_numbers[mask,] = nimg_new
            else:
                submasks[mask, :] = np.logical_and(
                    submasks[mask, :],
                    is_static
                )

            if prev_step is not None:
                prev_step = np.zeros_like(guesses)

        done = np.where(
            np.logical_not(np.any(submasks[mask,], axis=1))
        )[0]
        if len(done) > 0:  # easy check
            mask = np.delete(mask, done)
            if len(mask) == 0:
                break
        its[mask,] += 1
    else:
        converged = False
        its[mask,] = max_iterations

    guesses = guesses.reshape(base_shape + guesses.shape[-2:])
    errs = errs.reshape(base_shape)
    its = its.reshape(base_shape)

    if unitary and generate_rotation:
        raise NotImplementedError(...)
        res = (guess, rotations), converged, (errs, its)
    else:
        res = (guesses, image_numbers), converged, (errs, its)

    return res

class Damper:
    def __init__(self, damping_parameter=None, damping_exponent=None, restart_interval=10):
        self.n = 0
        self.u = damping_parameter
        self.exp = 1.0 if damping_exponent is None else damping_exponent
        self.restart = restart_interval

    def get_damping_factor(self):
        u = self.u
        if u is not None:
            if self.exp > 0:
                u = np.power(u, self.n*self.exp)
                self.n = (self.n + 1) % self.restart
        return u

class LineSearcher(metaclass=abc.ABCMeta):
    """
    Adapted from scipy.optimize to handle multiple structures at once
    """

    def __init__(self, func, min_alpha=0, **opts):
        self.func = func
        self.opts = opts
        self.min_alpha = min_alpha

    @abc.abstractmethod
    def check_scalar_converged(self, phi_vals, alphas, **opts):
        raise NotImplementedError("abstract")

    @abc.abstractmethod
    def update_alphas(self,
                      phi_vals, alphas, iteration,
                      old_phi_vals, old_alphas_vals,
                      mask,
                      **opts
                      ):
        raise NotImplementedError("abstract")

    default_alpha = 1e-3
    def get_default_alpha(self, am):
        return np.full_like(am, self.default_alpha)
    def scalar_search(self,
                      scalar_func,
                      guess_alpha,
                      min_alpha=None,
                      max_iterations=15,
                      history_length=1,
                      **opts):
        if min_alpha is None:
            min_alpha = self.min_alpha

        alphas = np.asanyarray(guess_alpha)
        mask = np.arange(len(alphas))

        phi_vals = scalar_func(alphas, mask)
        if history_length > 0:
            history = collections.deque(maxlen=history_length)
        else:
            history = None

        is_converged = np.full(len(mask), False)
        converged = np.where(self.check_scalar_converged(phi_vals, alphas, **opts))[0]
        if len(converged) > 0:
            is_converged[converged,] = True
            mask = np.delete(mask, converged)
            if len(mask) == 0:
                return alphas,  (phi_vals, is_converged)

        for i in range(max_iterations):
            if history is not None:
                phi_vals_old = [p[mask,] for p,a in history]
                alpha_vals_old = [a[mask,] for p,a in history]
            else:
                phi_vals_old = None
                alpha_vals_old = None

            new_alphas = self.update_alphas(phi_vals, alphas, i,
                                            phi_vals_old, alpha_vals_old,
                                            mask,
                                            **opts
                                            )
                # mask = np.delete(mask, problem_alphas)
                # if len(mask) == 0:
                #     break  # alphas, (phi_vals, is_converged)

            history.append([phi_vals.copy(), alphas.copy()])

            new_phi = scalar_func(new_alphas, mask)
            phi_vals[mask,] = new_phi
            # prev_alphas = alphas[mask,].copy()
            alphas[mask,] = new_alphas

            problem_alphas = np.where(new_alphas < min_alpha)[0]
            if len(problem_alphas) > 0:
                alphas[mask[problem_alphas,],] = min_alpha
                new_alphas = np.delete(new_alphas, problem_alphas)
                new_phi = np.delete(new_phi, problem_alphas)
                mask = np.delete(mask, problem_alphas)
                if len(mask) == 0:
                    break

            converged = np.where(self.check_scalar_converged(new_phi, new_alphas, **opts))[0]
            if len(converged) > 0:
                is_converged[mask[converged,],] = True
                mask = np.delete(mask, converged)
                if len(mask) == 0:
                    break

            # problem_alphas = np.where(np.abs(prev_alphas - alphas[mask,]) < 1e-8)[0]
            # if len(problem_alphas) > 0:
            #     mask = np.delete(mask, problem_alphas)
            #     if len(mask) == 0:
            #         break# alphas, (phi_vals, is_converged)
        else:
            am = alphas[mask,]
            default_alpha = self.get_default_alpha(am, **opts)
            alphas[mask,] = np.min(np.array([am, default_alpha]), axis=0)
        return alphas, (phi_vals, is_converged)

    def prep_search(self, initial_geom, search_dir, guess_alpha=1, **opts):
        return np.full(len(initial_geom), guess_alpha), opts, self._dir_func(self.func, initial_geom, search_dir)

    @classmethod
    def _dir_func(cls, func, initial_geom, search_dir):
        @functools.wraps(func)
        def phi(alphas, mask):
            return func(initial_geom[mask,] + alphas[:, np.newaxis] * search_dir[mask,], mask)

        return phi

    def __call__(self, initial_geom, search_dir, **base_opts):
        opts = dict(self.opts, **base_opts)
        guess_alpha, opts, phi = self.prep_search(initial_geom, search_dir, **opts)
        conv = self.scalar_search(
            phi,
            guess_alpha,
            **opts
        )
        return conv

class ArmijoSearch(LineSearcher):

    def __init__(self, func, c1=1e-4, min_alpha=None, fixed_step_cutoff=1e-8, der_max=1e2, guess_alpha=1):
        super().__init__(func, min_alpha=min_alpha, c1=c1)
        self.func = func
        self.der_max = der_max
        self.fixed_step_cutoff = fixed_step_cutoff
        self.guess_alpha = guess_alpha

    def prep_search(self, initial_geom, search_dir, *, initial_grad, min_alpha=None, **rest):
        mask = np.arange(len(initial_geom))
        derphi0 = np.reshape(initial_grad[:, np.newaxis, :] @ search_dir[:, :, np.newaxis], (-1,))
        derphi0 = np.clip(derphi0, -self.der_max, self.der_max)
        if min_alpha is None:
            min_alpha = self.min_alpha

        a0, opts, phi = super().prep_search(initial_geom, search_dir, **rest)
        if self.fixed_step_cutoff is None:
           if min_alpha is None:
               min_alpha = 1e-8
        else:
           if min_alpha is None:
               min_alpha = 1e-8 #if np.max(np.abs(derphi0)) > self.fixed_step_cutoff else 1
           a0 = (
                   (np.abs(derphi0) > self.fixed_step_cutoff) + (np.abs(derphi0) <= self.fixed_step_cutoff)
           ).astype(float)

        phi0 = phi(np.zeros_like(a0), mask)
        return a0, dict(opts, phi0=phi0, derphi0=derphi0, min_alpha=min_alpha), phi

    converged_tolerance = 1e-8
    def check_scalar_converged(self, phi_vals, alphas, *, phi0, c1, derphi0, tol=None):
        if tol is None:
            tol = self.converged_tolerance
        test = phi0 + c1 * alphas * derphi0
        return np.logical_and(
            np.logical_not(np.isnan(phi_vals)),
            np.logical_or(
                phi_vals < test,
                np.allclose(phi_vals, test, rtol=0, atol=tol)
            )
        )

    def get_default_alpha(self, am, *, phi0, **etc):
        return np.full_like(am, self.default_alpha / np.abs(phi0))

    def update_alphas(self,
                      phi_vals, alphas, iteration,
                      old_phi_vals, old_alphas_vals,
                      mask,
                      *,
                      phi0, c1, derphi0,
                      zero_cutoff=1e-16
                      ):
        phi0 = phi0[mask,]
        derphi0 = derphi0[mask,]

        if iteration == 0:
            factor = (phi_vals - phi0 - derphi0 * alphas)
            # alpha1 = alphas.copy()
            # safe_pos = np.where(np.abs(factor) > zero_cutoff)
            # alpha1[safe_pos,] = -(derphi0[safe_pos,]) * alphas[safe_pos] ** 2 / 2.0 / factor[safe_pos]
            # TODO: ensure stays numerically stable
            # print(".>>", phi0)
            alpha1 = -(derphi0) * alphas ** 2 / 2.0 / factor
            alpha_new = alpha1
        else:
            phi_a0 = old_phi_vals[0]
            phi_a1 = phi_vals
            alpha0 = old_alphas_vals[0]
            alpha1 = alphas

            # da = (alpha1 - alpha0)

            # safe_pos = np.where(np.abs(factor) < zero_cutoff)
            # factor = alpha0 ** 2 * alpha1 ** 2 * (alpha1 - alpha0)
            # a = alpha0 ** 2 * (phi_a1 - phi0 - derphi0 * alpha1) - \
            #     alpha1 ** 2 * (phi_a0 - phi0 - derphi0 * alpha0)
            # a = a / factor
            # b = -alpha0 ** 3 * (phi_a1 - phi0 - derphi0 * alpha1) + \
            #     alpha1 ** 3 * (phi_a0 - phi0 - derphi0 * alpha0)
            # b = b / factor

            # scaling = 1
            n0 = (phi_a0 - phi0 - derphi0 * alpha0)
            n1 = (phi_a1 - phi0 - derphi0 * alpha1)
            d0 = alpha0 ** 2 * (alpha1 - alpha0)
            d1 = alpha1 ** 2 * (alpha1 - alpha0)
            # if d0 < 1e-8 or d1 < 1e-8:
            #     scaling = 1e6
            #     d0 = d0 * scaling
            #     d1 = d1 * scaling
            f1 = n1 / d1
            f0 = n0 / d0

            a = f1 - f0
            b = alpha1 * f0 - alpha0 * f1

            alpha2 = (-b + np.sqrt(abs(b ** 2 - 3 * a * derphi0))) / (3.0 * a)
            # alpha2 = alpha2 / scaling

            halved_alphas = np.where(
                np.logical_or(
                    (alpha1 - alpha2) > alpha1 / 2.0,
                    (1 - alpha2 / alpha1) < 0.96
                )
            )
            alpha2[halved_alphas] = alpha1[halved_alphas] / 2.0

            alpha_new = alpha2

        bad_pos = np.where(np.isnan(alpha_new))
        alpha_new[bad_pos] = alphas[bad_pos] / 2
        return alpha_new

class _WolfeLineSearch(LineSearcher):
    """
    Adapted from scipy.optimize
    """

    def __init__(self, func, grad, **opts):
        super().__init__(func)
        self.grad = grad

    @classmethod
    def _grad_func(cls, jac, initial_geom, search_dir):
        @functools.wraps(jac)
        def derphi(alphas, mask):
            pk = search_dir[mask,]
            grad = jac(initial_geom[mask,] + alphas[:, np.newaxis] * pk, mask)
            return (grad[: np.newaxis, :] @ pk[:, :, np.newaxis]).reshape(-1)

        return derphi

    def prep_search(self, initial_geom, search_dir):
        raise NotImplementedError()
        a_guess, opts, phi = super().prep_search(initial_geom, search_dir)
        derphi = self._grad_func(self.grad, initial_geom, search_dir)
        gfk = self.grad(initial_geom)
        derphi0 = derphi(np.zeros_like(a_guess), np.arange(len(a_guess)))

        # stp, fval, old_fval = scalar_search_wolfe1(
        #     phi, derphi, old_fval, old_old_fval, derphi0,
        #     c1=c1, c2=c2, amax=amax, amin=amin, xtol=xtol)

        # return stp, fc[0], gc[0], fval, old_fval, gval[0]

    def check_scalar_converged(self, phi_vals, alphas, **opts):
        ...

class GradientDescentStepFinder:
    supports_hessian = False
    line_search = ArmijoSearch

    def __init__(self, func, jacobian, damping_parameter=None, damping_exponent=None,
                 line_search=True, restart_interval=10, logger=None):
        self.func = func
        self.jac = jacobian
        self.damper = Damper(
            damping_parameter=damping_parameter,
            damping_exponent=damping_exponent,
            restart_interval=restart_interval
        )

        if line_search is True:
            line_search = self.line_search(func)
        elif line_search is False:
            line_search = None
        self.searcher = line_search
        self.logger = logger

    def __call__(self, guess, mask, projector=None):
        if isinstance(mask, tuple):  # for chain minimizers
            mask, (j, _, _) = mask
            guess = guess[:, j]

        jacobian = self.jac(guess, mask)

        new_step_dir = -jacobian
        if projector is not None:
            new_step_dir = (new_step_dir[..., np.newaxis, :] @ projector).reshape(new_step_dir.shape)

        if self.searcher is not None:
            alpha, (fvals, is_converged) = self.searcher(guess, new_step_dir, initial_grad=jacobian)
            new_step_dir = alpha[:, np.newaxis] * new_step_dir

        # h = func(guess, mask)
        u = self.damper.get_damping_factor()
        if u is not None:
            new_step_dir = new_step_dir * u

        return new_step_dir, jacobian

class NetwonDirectHessianGenerator:

    line_search = ArmijoSearch
    def __init__(self, func, jacobian, hessian, hess_mode='direct', line_search=True,
                 damping_parameter=None, damping_exponent=None, restart_interval=10
                 ):
        hessian = self.wrap_hessian(hessian, hess_mode)
        self.jacobian = jacobian
        self.hessian_inverse = hessian
        self.damper = Damper(
            damping_parameter=damping_parameter,
            damping_exponent=damping_exponent,
            restart_interval=restart_interval
        )
        if line_search is True:
            line_search = self.line_search(func)
        elif line_search is False:
            line_search = None
        self.searcher = line_search

    def wrap_hessian(self, func, mode):
        if mode == 'direct':
            @functools.wraps(func)
            def hessian_inverse(guess, mask):
                h = func(guess, mask)
                # u = self.damper.get_damping_factor()
                # if u is not None:
                #     h = h + u * vec_ops.identity_tensors(guess.shape[:-1], guess.shape[-1])
                return np.linalg.inv(h)
        else:
            hessian_inverse = func
            # @functools.wraps(func)
            # def hessian_inverse(guess, mask):
            #     return h

        return hessian_inverse

    def __call__(self, guess, mask, projector=None):

        jacobian, hessian_inv = self.jacobian(guess, mask), self.hessian_inverse(guess, mask)

        new_step_dir = -(hessian_inv @ jacobian[:, :, np.newaxis]).reshape(jacobian.shape)
        if projector is not None:
            new_step_dir = (new_step_dir[:, np.newaxis, :] @ projector).reshape(new_step_dir.shape)
        if self.searcher is not None:
            alpha, (fvals, is_converged) = self.searcher(guess, new_step_dir, initial_grad=jacobian)
            new_step_dir = alpha[:, np.newaxis] * new_step_dir

        # h = func(guess, mask)
        u = self.damper.get_damping_factor()
        if u is not None:
            new_step_dir = new_step_dir * u

        return new_step_dir, jacobian

class NewtonStepFinder:
    supports_hessian = True
    def __init__(self, func, jacobian=None, hessian=None, *, check_generator=True, logger=None, **generator_opts):
        if check_generator:
            generator = self._prep_generator(func, jacobian, hessian, generator_opts)
        else:
            generator = jacobian
        self.generator = generator
        self.logger = logger

    @classmethod
    def _prep_generator(cls, func, jac, hess, opts):
        if (hasattr(func, 'jacobian') and hasattr(func, 'hessian_inverse')):
            return func
        else:
            if hess is None:
                raise ValueError(
                    "Direct Netwon requires a Hessian or a generator for the Jacobian and Hessian inverse. "
                    "Consider using Quasi-Newton if only the Jacobian is fast to compute.")
            return NetwonDirectHessianGenerator(func, jac, hess, **opts)

    def __call__(self, guess, mask, projector=None):
        if isinstance(mask, tuple):  # for chain minimizers
            mask, (j, _, _) = mask
            guess = guess[:, j]
        return self.generator(guess, mask, projector=projector)

class QuasiNewtonStepFinder:
    supports_hessian = False

    def __init__(self, func, jacobian, approximation_type='bfgs', logger=None, **generator_opts):
        self.hess_appx = self.hessian_approximations[approximation_type.lower()](func, jacobian, **generator_opts)
    @classmethod
    def get_hessian_approximations(cls):
        return {
            'bfgs': BFGSApproximator,
            'broyden': BroydenApproximator,
            'dfp': DFPApproximator,
            'sr1': SR1Approximator,
            'psb': PSBQuasiNewtonApproximator,
            'bofill': BofillApproximator,
            'schelegel': SchelgelApproximator,
            'greenstadt': GreenstadtNewtonApproximator
        }
    @property
    def hessian_approximations(self):
        return self.get_hessian_approximations()

    def __call__(self, guess, mask, projector=None):
        if isinstance(mask, tuple):  # for chain minimizers
            mask, (j, _, _) = mask
            guess = guess[:, j]
        return self.hess_appx(guess, mask, projector=projector)

class QuasiNetwonHessianApproximator:
    orthogonal_dirs_cutoff = 1e-16#1e-8
    line_search = ArmijoSearch
    def __init__(self, func, jacobian, initial_beta=1,
                 damping_parameter=None, damping_exponent=None,
                 line_search=True, restart_interval=10,
                 restart_hessian_norm=1e-12,
                 # approximation_mode='direct'
                 approximation_mode='inverse'
                 ):
        self.func = func
        self.jac = jacobian
        self.initial_beta = initial_beta
        self.base_hess = None
        self.prev_jac = None
        self.prev_step = None
        self.prev_hess_inv = None
        self.eye_tensors = None
        self.damper = Damper(
            damping_parameter=damping_parameter,
            damping_exponent=damping_exponent,
            restart_interval=restart_interval
        )
        if line_search is True:
            line_search = self.line_search(func)
        elif line_search is False:
            line_search = None
        self.searcher = line_search
        self.restart_hessian_norm = restart_hessian_norm
        self.approximation_mode = approximation_mode

    def identities(self, guess, mask):
        if self.eye_tensors is None:
            self.eye_tensors = vec_ops.identity_tensors(guess.shape[:-1], guess.shape[-1])
            return self.eye_tensors
        else:
            return self.eye_tensors[mask,]

    def initialize_hessians(self, guess, mask):
        if self.approximation_mode == 'direct':
            return self.initial_beta * self.identities(guess, mask)
        else:
            return (1/self.initial_beta) * self.identities(guess, mask)

    @classmethod
    def take_nonzero_norm_regions(cls, norms, tensors, cutoff=None):
        if cutoff is None:
            cutoff = cls.orthogonal_dirs_cutoff
        mask = np.full(norms[0].reshape(-1,).shape, True)
        if cutoff is not None:
            for n in norms:
                mask = np.logical_and(mask, np.abs(n.reshape(-1)) > cutoff)

        # good_pos = np.where(np.logical_and(*[
        #     np.abs(n.reshape(-1, )) > cutoff
        #     for n in norms
        # ]))
        good_pos = np.where(mask)
        return good_pos, [t[good_pos] for t in tensors]

    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        raise NotImplementedError("abstract")

    def get_jacobian_updates(self, guess, mask):
        new_jacs = self.jac(guess, mask)
        if self.prev_jac is None:
            jac_diffs = new_jacs
        else:
            prev_jacs = self.prev_jac[mask,]
            jac_diffs = new_jacs - prev_jacs
        return new_jacs, jac_diffs

    def restart_hessian_approximation(self):
        restart = np.any(
            np.linalg.norm(self.prev_step, axis=-1) < self.restart_hessian_norm
        )
        return restart

    def __call__(self, guess, mask, projector=None):
        new_jacs, jacobian_diffs = self.get_jacobian_updates(guess, mask)
        if self.prev_step is None or self.restart_hessian_approximation():
            new_hess = self.initialize_hessians(guess, mask)
        else:
            prev_steps = self.prev_step[mask,]
            prev_hess = self.prev_hess_inv[mask,]
            new_hess = self.get_hessian_update(self.identities(guess, mask), jacobian_diffs, prev_steps, prev_hess)

        if self.approximation_mode == 'direct':
            B = np.linalg.inv(new_hess)
        else:
            B = new_hess
        new_step_dir = -(B @ new_jacs[:, :, np.newaxis]).reshape(new_jacs.shape)
        if projector is not None:
            new_step_dir = (new_step_dir[:, np.newaxis, :] @ projector).reshape(new_step_dir.shape)
        if self.searcher is not None:
            alpha, (fvals, is_converged) = self.searcher(guess, new_step_dir, initial_grad=new_jacs)
        else:
            alpha = np.ones(len(new_step_dir))
        # handle convergence issues?
        new_step = alpha[:, np.newaxis] * new_step_dir
        # print(np.isnan(guess).any(), np.isnan(alpha).any(), np.linalg.norm(new_step_dir))
        u = self.damper.get_damping_factor()
        if u is not None:
            new_step = new_step * u

        if self.prev_jac is None:
            self.prev_jac = new_jacs
        else:
            self.prev_jac[mask,] = new_jacs

        if self.prev_step is None:
            self.prev_step = new_step
        else:
            self.prev_step[mask,] = new_step

        if self.prev_hess_inv is None:
            self.prev_hess_inv = new_hess
        else:
            self.prev_hess_inv[mask,] = new_hess

        return new_step, new_jacs

class BFGSApproximator(QuasiNetwonHessianApproximator):

    orthogonal_dirs_cutoff = 1e-16 #1e-8
    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        I = identities
        dx = prev_steps[:, :, np.newaxis]
        dx_T = prev_steps[:, np.newaxis, :]
        y = jacobian_diffs[:, :, np.newaxis]
        y_T = jacobian_diffs[:, np.newaxis, :]
        B = prev_hess.copy()
        increment = False
        if self.approximation_mode == 'direct':
            diff_norm = y_T @ dx
            h_step = (B @ dx)
            h_norm = dx_T @ h_step
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                h_step, h_norm, diff_norm
            ) = self.take_nonzero_norm_regions([diff_norm, h_norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                h_step, h_norm, diff_norm])
            h_step_T = np.moveaxis(h_step, -1, -2)

            diff_outer = y_T * y
            step_outer = h_step_T * h_step
            diff_step = diff_outer / diff_norm
            step_step = step_outer / h_norm
            update = diff_step - step_step
            increment = True
        else:
            diff_norm = y_T @ dx
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                diff_norm
            ) = self.take_nonzero_norm_regions([diff_norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                diff_norm])

            diff_outer = np.moveaxis(dx_T * y, -1, -2)
            diff_step = I - diff_outer / diff_norm
            step_outer = dx_T * dx
            step_step = step_outer / diff_norm
            update = diff_step @ H @ np.moveaxis(diff_step, -1, -2) + step_step
            increment = False

        if increment:
            B[good_pos] += update
        else:
            B[good_pos] = update
        return B

class DFPApproximator(QuasiNetwonHessianApproximator):

    orthogonal_dirs_cutoff = 1e-8
    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        I = identities
        dx = prev_steps[:, :, np.newaxis]
        dx_T = prev_steps[:, np.newaxis, :]
        y = jacobian_diffs[:, :, np.newaxis]
        y_T = jacobian_diffs[:, np.newaxis, :]
        B = prev_hess.copy()
        if self.approximation_mode == 'direct':
            norm = y_T @ dx
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                norm
            ) = self.take_nonzero_norm_regions([norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                norm])

            proj = I - (y @ dx_T) / norm
            update = proj @ H[good_pos] @ np.moveaxis(proj, -1, -2)

            update = update + (y @ y_T)/norm
        else:
            norm = y_T @ dx
            h_step = B @ y
            h_norm = y_T @ h_step
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                norm, h_step, h_norm
            ) = self.take_nonzero_norm_regions([norm, h_norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                norm, h_step, h_norm])

            update = dx @ dx_T - (h_step/h_norm) @ np.moveaxis(h_step, -1, -2)

        B[good_pos] += update
        return B

class BroydenApproximator(QuasiNetwonHessianApproximator):

    orthogonal_dirs_cutoff = 1e-8
    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        I = identities
        dx = prev_steps[:, :, np.newaxis]
        dx_T = prev_steps[:, np.newaxis, :]
        y = jacobian_diffs[:, :, np.newaxis]
        y_T = jacobian_diffs[:, np.newaxis, :]
        B = prev_hess.copy()
        if self.approximation_mode == 'direct':
            dx_norm = dx_T * dx
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                dx_norm
            ) = self.take_nonzero_norm_regions([dx_norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                dx_norm])

            h_step = (H @ dx)
            update = (y - h_step)/dx_norm * dx_T

            B[good_pos] += update * dx_T
        else:
            h_y = B @ y
            h_x = B @ dx
            h_norm = dx_T @ h_y
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                h_y, h_x, h_norm
            ) = self.take_nonzero_norm_regions([h_norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                h_y, h_x, h_norm])

            h_step = (dx - h_y) / h_norm
            update = h_step * np.moveaxis(h_x, -1, -2)

        B[good_pos] += update
        return B

class SR1Approximator(QuasiNetwonHessianApproximator):

    orthogonal_dirs_cutoff = 1e-8
    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        I = identities
        dx = prev_steps[:, :, np.newaxis]
        dx_T = prev_steps[:, np.newaxis, :]
        y = jacobian_diffs[:, :, np.newaxis]
        y_T = jacobian_diffs[:, np.newaxis, :]
        B = prev_hess.copy()
        if self.approximation_mode == 'direct':
            h_step = y - (B @ dx)
            h_step_T = np.moveaxis(h_step, -1, -2)
            h_norm = dx_T @ h_step
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                h_step, h_step_T, h_norm
            ) = self.take_nonzero_norm_regions([h_norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                h_step, h_step_T, h_norm])
            update = (h_step_T * h_step) / h_norm
        else:
            h_step = dx - (B @ y)
            h_step_T = np.moveaxis(h_step, -1, -2)
            h_norm = y_T @ h_step
            good_pos, (
                I, H, dx, dx_T, y, y_T,
                h_step, h_step_T, h_norm
            ) = self.take_nonzero_norm_regions([h_norm],
                                               [I, B, dx, dx_T, y, y_T,
                                                h_step, h_step_T, h_norm])
            update = (h_step_T * h_step) / h_norm

        B[good_pos] += update
        return B

class CompactQuasiNewtonApproximator(QuasiNetwonHessianApproximator):
    @classmethod
    def get_direct_hessian_update_vector(cls, H, dx, y):
        raise NotImplementedError("abstract")
    @classmethod
    def get_inverse_hessian_update_vector(cls, H, dx, y):
        raise NotImplementedError("abstract")

    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        I = identities
        dx = prev_steps[:, :, np.newaxis]
        dx_T = prev_steps[:, np.newaxis, :]
        y = jacobian_diffs[:, :, np.newaxis]
        y_T = jacobian_diffs[:, np.newaxis, :]
        B = prev_hess.copy()

        if self.approximation_mode == 'direct':
            v = self.get_direct_hessian_update_vector(
                B, dx, y
            )
        else:
            v = self.get_inverse_hessian_update_vector(
                B, dx, y
            )
            y, y_T, dx, dx_T = dx, dx_T, y, y_T

        norm = v @ dx_T
        good_pos, (
            I, H, dx, dx_T, y, y_T,
            v, norm
        ) = self.take_nonzero_norm_regions([norm],
                                           [I, B, dx, dx_T, y, y_T,
                                            v, norm])

        d = (dx - H @ y) / norm
        d_T = np.moveaxis(d, -1, -2)
        v_T = np.moveaxis(v, -1, -2)
        dv = d * v_T
        dv_T = np.moveaxis(dv, -1, -2)

        B[good_pos,] += dv + dv_T - ((d_T @ y) / norm) * (v * v_T)

        return B

class PSBQuasiNewtonApproximator(CompactQuasiNewtonApproximator):
    @classmethod
    def get_direct_hessian_update_vector(cls, H, dx, y):
        raise NotImplementedError("PSB only does inverse mode")
    @classmethod
    def get_inverse_hessian_update_vector(cls, H, dx, y):
        return dx

class GreenstadtNewtonApproximator(CompactQuasiNewtonApproximator):
    @classmethod
    def get_direct_hessian_update_vector(cls, H, dx, y):
        raise y

    @classmethod
    def get_inverse_hessian_update_vector(cls, H, dx, y):
        raise NotImplementedError("Greenstadt only does direct mode")

class WeightedQuasiNewtonApproximator(QuasiNetwonHessianApproximator):

    base_approximators = None
    def __init__(
            self,
            func, jacobian, initial_beta=1,
            damping_parameter=None, damping_exponent=None,
            line_search=True, restart_interval=10,
            restart_hessian_norm=1e-5,
            approximation_mode='direct'
    ):
        super().__init__(
            func, jacobian,
            initial_beta=initial_beta,
            damping_parameter=damping_parameter, damping_exponent=damping_exponent,
            line_search=line_search, restart_interval=restart_interval,
            restart_hessian_norm=restart_hessian_norm,
            approximation_mode=approximation_mode
        )
        self.approximators = [
            app(
                None, None,
                initial_beta=initial_beta,
                damping_parameter=damping_parameter, damping_exponent=damping_exponent,
                line_search=line_search, restart_interval=restart_interval,
                restart_hessian_norm=restart_hessian_norm,
                approximation_mode=approximation_mode
            )
            for app in self.base_approximators
        ]

    def get_direct_weights(self, jacobian_diffs, prev_steps, prev_hess):
        raise NotImplementedError("abstract")
    def get_inverse_weights(self, jacobian_diffs, prev_steps, prev_hess):
        raise NotImplementedError("abstract")


    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        if self.approximation_mode == 'direct':
            weights = self.get_direct_weights(jacobian_diffs, prev_steps, prev_hess)
        else:
            weights = self.get_inverse_weights(jacobian_diffs, prev_steps, prev_hess)

        base_updates = np.array([
            app.get_hessian_update(identities, jacobian_diffs, prev_steps, prev_hess)
            for app in self.approximators
        ])

        # weights is kxb array and base_updates is bxkxnxn
        avg = weights[:, np.newaxis, :] @ np.moveaxis(base_updates, 1, 0)[:, :, np.newaxis, :, :]
        return avg.reshape(base_updates.shape[1:])

class BofillApproximator(WeightedQuasiNewtonApproximator):
    base_approximators = [
        SR1Approximator,
        PSBQuasiNewtonApproximator
    ]
    def get_direct_weights(self, jacobian_diffs, prev_steps, prev_hess):
        raise NotImplementedError("only inverse supported")

    @classmethod
    def get_psi(self, jacobian_diffs, prev_steps, prev_hess):
        psi = np.ones_like(prev_steps)

        dx = prev_steps[:, :, np.newaxis]
        dx_T = prev_steps[:, np.newaxis, :]
        y = jacobian_diffs[:, :, np.newaxis]
        y_T = jacobian_diffs[:, np.newaxis, :]
        H = prev_hess
        h_step = y + H @ dx
        h_step_T = np.moveaxis(h_step, -2, -1)
        h_norm = h_step_T @ h_step
        s_norm = dx_T @ dx
        good_pos, (
            H, dx, dx_T, y, y_T,
            s_norm, h_norm
        ) = self.take_nonzero_norm_regions([s_norm, h_norm],
                                           [H, dx, dx_T, y, y_T,
                                            s_norm, h_norm])
        step_disp = h_step_T @ dx

        psi[good_pos] = np.reshape( (step_disp**2)/(s_norm * h_norm), -1)

        return psi

    def get_inverse_weights(self, jacobian_diffs, prev_steps, prev_hess):
            psi = self.get_psi(jacobian_diffs, prev_steps, prev_hess)
            return [psi, 1-psi]

class SchelgelApproximator(BofillApproximator):
    base_approximators = [
        SR1Approximator,
        BFGSApproximator
    ]
    def get_direct_weights(self, jacobian_diffs, prev_steps, prev_hess):
        raise NotImplementedError("only inverse supported")

    def get_psi(self, jacobian_diffs, prev_steps, prev_hess):
        psi = np.sqrt(BofillApproximator.get_psi())

        return psi

    def get_inverse_weights(self, jacobian_diffs, prev_steps, prev_hess):
            psi = np.sqrt(BofillApproximator.get_psi(jacobian_diffs, prev_steps, prev_hess))
            return [psi, 1-psi]

class ConjugateGradientStepFinder:
    supports_hessian = False

    def __init__(self, func, jacobian, approximation_type='polak-ribiere', logger=None, **generator_opts):
        self.step_appx = self.beta_approximations[approximation_type.lower()](func, jacobian, **generator_opts)
        self.logger = logger
    @property
    def beta_approximations(self):
        return {
            'fletcher-reeves':FletcherReevesApproximator,
            'polak-ribiere':PolakRibiereApproximator
        }

    def __call__(self, guess, mask, projector=None):
        if isinstance(mask, tuple):  # for chain minimizers
            mask, (j, _, _) = mask
            guess = guess[:, j]
        return self.step_appx(guess, mask, projector=projector)

class ConjugateGradientStepApproximator:
    line_search = ArmijoSearch

    def __init__(self, func, jacobian,
                 damping_parameter=None, damping_exponent=None,
                 restart_interval=50, restart_parameter=0.9,
                 line_search=True):
        self.func = func
        self.jac = jacobian
        self.base_hess = None
        self.prev_jac = None
        self.prev_step_dir = None
        self.damper = Damper(
            damping_parameter=damping_parameter,
            damping_exponent=damping_exponent,
            restart_interval=restart_interval
        )
        self.n = 0
        self.restart_interval = restart_interval
        self.restart_parameter = restart_parameter
        if line_search is True:
            line_search = self.line_search(func)
        elif line_search is False:
            line_search = None
        self.searcher = line_search

    def get_beta(self, new_jacs, prev_jac, prev_step_dir):
        raise NotImplementedError("abstract")

    def determine_restart(self, new_jacs, mask):
        if self.n == 0: return True
        if self.restart_parameter is not None:
            prev_jac = self.prev_jac[mask,]
            new_norm = np.abs(new_jacs[:, np.newaxis, :] @ prev_jac[:, :, np.newaxis]).flatten()
            old_norm = (prev_jac[:, np.newaxis, :] @ prev_jac[:, :, np.newaxis]).flatten()
            return np.any(
                self.restart_parameter * old_norm < new_norm
            )

    def __call__(self, guess, mask, projector=None):
        new_jacs = self.jac(guess, mask)

        if self.prev_jac is None or self.determine_restart(new_jacs, mask):
            new_step_dir = -new_jacs
        else:
            prev_jac = self.prev_jac[mask,]
            prev_step_dir = self.prev_step_dir[mask,]
            beta = self.get_beta(new_jacs, prev_jac, prev_step_dir)
            new_step_dir = -new_jacs + beta[:, np.newaxis] * prev_step_dir

        if projector is not None:
            new_step_dir = (new_step_dir[:, np.newaxis, :] @ projector).reshape(new_step_dir.shape)

        if self.searcher is not None:
            alpha, (fvals, is_converged) = self.searcher(guess, new_step_dir, initial_grad=new_jacs)
        else:
            alpha = np.ones(len(new_step_dir))
        # handle convergence issues?
        new_step = alpha[:, np.newaxis] * new_step_dir

        u = self.damper.get_damping_factor()
        if u is not None:
            new_step = new_step * u

        if self.prev_jac is None:
            self.prev_jac = new_jacs
        else:
            self.prev_jac[mask,] = new_jacs

        if self.prev_step_dir is None:
            self.prev_step_dir = new_step_dir
        else:
            self.prev_step_dir[mask,] = new_step_dir

        self.n = (self.n + 1) % self.restart_interval

        return new_step, new_jacs

class FletcherReevesApproximator(ConjugateGradientStepApproximator):
    def get_beta(self, new_jacs, prev_jac, prev_step_dir):
        return (
                (new_jacs[:, np.newaxis, :] @ new_jacs[:, :, np.newaxis]) /
                (prev_jac[:, np.newaxis, :] @ prev_jac[:, :, np.newaxis])
        ).reshape(len(new_jacs))

class PolakRibiereApproximator(ConjugateGradientStepApproximator):
    def get_beta(self, new_jacs, prev_jac, prev_step_dir):
        return (
                (new_jacs[:, np.newaxis, :] @ (new_jacs[:, :, np.newaxis] - prev_jac[:, :, np.newaxis])) /
                (prev_jac[:, np.newaxis, :] @ prev_jac[:, :, np.newaxis])
        ).reshape(len(new_jacs))

class EigenvalueFollowingStepFinder:

    line_search = ArmijoSearch
    def __init__(self, func, jacobian, hessian, initial_beta=1,
                 damping_parameter=None, damping_exponent=None,
                 line_search=False, restart_interval=1,
                 restart_hessian_norm=1e-5,
                 hessian_approximator='bofill',
                 approximation_mode='direct',
                 target_mode=None,
                 logger=None
                 # approximation_mode='inverse'
                 ):
        self.base_approximator = QuasiNewtonStepFinder.get_hessian_approximations()[hessian_approximator.lower()](
            func, jacobian
        )
        self.logger = logger

        self.func = func
        self.jac = jacobian
        self.hess = hessian
        self.initial_beta = initial_beta
        self.base_hess = None
        self.prev_jac = None
        self.prev_step = None
        self.prev_evec = None
        self.prev_hess = None
        self.eye_tensors = None
        self.damper = Damper(
            damping_parameter=damping_parameter,
            damping_exponent=damping_exponent,
            restart_interval=restart_interval
        )
        if line_search is True:
            line_search = self.line_search(func)
        elif line_search is False:
            line_search = None
        self.searcher = line_search
        self.restart_hessian_norm = restart_hessian_norm
        self.approximation_mode = approximation_mode
        self.target_mode = target_mode

    def identities(self, guess, mask):
        if self.eye_tensors is None:
            self.eye_tensors = vec_ops.identity_tensors(guess.shape[:-1], guess.shape[-1])
            return self.eye_tensors
        else:
            return self.eye_tensors[mask,]

    def initialize_hessians(self, guess, mask):
        return self.hess(guess)

    def get_hessian_update(self, identities, jacobian_diffs, prev_steps, prev_hess):
        return self.base_approximator.get_hessian_update(identities, jacobian_diffs, prev_steps, prev_hess)

    negative_eigenvalue_offset = 0.015
    positive_eigenvalue_offset = 0.005
    mode_tracking_overlap_cutoff = 1e-5
    def get_shift(self, evals, tf_new, target_mode):
        if target_mode is None:
            target_ev = evals[0] # could do np.min(ev) if we had some other (or complex) eigensolver
        else:
            overlaps = np.abs(tf_new @ target_mode) # TODO: check if this makes sense, in principle under a
                                                    # quadratic appx. the "direction" of the step doesn't matter
            abs_ov = np.abs(overlaps)
            ev_pos = np.argmax(abs_ov)
            if abs_ov[ev_pos] < self.mode_tracking_overlap_cutoff:
                target_ev = evals[0]
            else:
                target_ev = evals[ev_pos]

        if target_ev < 0:
            shift = -target_ev + self.negative_eigenvalue_offset # make this very slightly positive
        else:
            shift = self.positive_eigenvalue_offset

        return shift

    def get_jacobian_updates(self, guess, mask):
        new_jacs = self.jac(guess, mask)
        if self.prev_jac is None:
            jac_diffs = new_jacs
        else:
            prev_jacs = self.prev_jac[mask,]
            jac_diffs = new_jacs - prev_jacs
        return new_jacs, jac_diffs

    def restart_hessian_approximation(self):
        restart = np.any(
            np.linalg.norm(self.prev_step, axis=-1) < self.restart_hessian_norm
        )
        return restart

    def __call__(self, guess, mask, projector=None):
        if isinstance(mask, tuple):  # for chain minimizers
            mask, (j, _, _) = mask
            guess = guess[:, j]

        new_jacs, jacobian_diffs = self.get_jacobian_updates(guess, mask)
        if self.prev_step is None or self.restart_hessian_approximation():
            new_hess = self.initialize_hessians(guess, mask)
        else:
            prev_steps = self.prev_step[mask,]
            prev_hess = self.prev_hess[mask,]
            new_hess = self.get_hessian_update(guess, self.identities(guess, mask), jacobian_diffs, prev_steps, prev_hess)

        if self.prev_evec is None:
            prev_evec = None
        else:
            prev_evec = self.prev_evec[mask,]

        evals, tf = np.linalg.eigh(new_hess)
        if misc.is_numeric(self.target_mode):
            self.target_mode = tf[:, self.target_mode]

        shift = self.get_shift(evals, tf, self.target_mode)

        tf_grad = -np.diag(
                np.moveaxis(tf, -1, -2) @ new_jacs[:, :, np.newaxis]
        ) / (evals[:, :, np.newaxis] + shift[:, np.newaxis, np.newaxis])
        new_step_dir = (tf_grad @ tf).reshape(new_jacs.shape)

        # print(new_hess)
        if projector is not None:
            new_step_dir = (new_step_dir[:, np.newaxis, :] @ projector).reshape(new_step_dir.shape)
        if self.searcher is not None:
            alpha, (fvals, is_converged) = self.searcher(guess, new_step_dir, initial_grad=new_jacs)
        else:
            alpha = np.ones(len(new_step_dir))
        # handle convergence issues?
        new_step = alpha[:, np.newaxis] * new_step_dir
        # print(np.isnan(guess).any(), np.isnan(alpha).any(), np.linalg.norm(new_step_dir))
        u = self.damper.get_damping_factor()
        if u is not None:
            new_step = new_step * u

        if self.prev_jac is None:
            self.prev_jac = new_jacs
        else:
            self.prev_jac[mask,] = new_jacs

        if self.prev_step is None:
            self.prev_step = new_step
        else:
            self.prev_step[mask,] = new_step

        if self.prev_hess is None:
            self.prev_hess = new_hess
        else:
            self.prev_hess[mask,] = new_hess

        return new_step, new_jacs

class ChainMinimizingStepFinder:
    supports_hessian = True
    def __init__(self,
                 func,
                 jacobian,
                 hessian=None,
                 step_finder='conjugate-gradient',
                 logger=None,
                 **opts
                 ):
        self.step_finder = get_step_finder({
            'method':step_finder,
            'func':self.wrap_func(func),
            'jacobian':self.wrap_jac(jacobian),
            'hessian':self.wrap_hess(hessian),
            **opts
        })
        self._mask_data = None
        self.logger = logger

    def adjust_jacobian(self, jac, guess, mask, cur, prev, next):
        return jac

    def adjust_hessian(self, hess, guess, mask, cur, prev, next):
        return hess

    @abc.abstractmethod
    def image_pairwise_contribution(self, guess, mask, cur, prev, next, order=0):
        raise NotImplementedError("abstract")

    def wrap_func(self, func):
        @functools.wraps(func)
        def wrapped_func(guess, mask):
            j, prev, next = self._mask_data
            return func(guess[:, j], mask) + self.image_pairwise_contribution(guess, mask, j, prev, next, order=0)
        return wrapped_func

    def wrap_jac(self, jac):
        @functools.wraps(jac)
        def wrapped_jac(guess, mask):
            j, prev, next = self._mask_data
            base_jac = self.adjust_jacobian(
                jac(guess[:, j], mask),
                guess, mask, j, prev, next
            )
            new_jac = self.image_pairwise_contribution(guess, mask, j, prev, next, order=1)
            # print(base_jac)
            # print(new_jac)
            # print("-"*20)
            return base_jac + new_jac
        return wrapped_jac

    def wrap_hess(self, hess):
        if hess is None:
            return hess
        @functools.wraps(hess)
        def wrapped_hess(guess, mask):
            j, prev, next = self._mask_data
            return self.adjust_hessian(
                hess(guess[:, j], mask),
                guess, mask, j, prev, next
            ) + self.image_pairwise_contribution(guess, mask, j, prev, next, order=2)
        return wrapped_hess

    def __call__(self, guess, mask, projector=None):
        mask, self._mask_data = mask
        return self.step_finder(guess, mask, projector=projector)

class NudgedElasticBandStepFinder(ChainMinimizingStepFinder):
    def __init__(self,
                 func,
                 jacobian,
                 hessian=None,
                 spring_constants=.1,
                 distance_function=None,
                 step_finder='gradient-descent',
                 logger=None,
                 **opts
                 ):
        self.image_potential = func
        super().__init__(func, jacobian, hessian=hessian, step_finder=step_finder, **opts)
        self.spring_constants = spring_constants
        self._spring_constants = None
        self.distance_function = distance_function
        self._last_tangent = None
        self.logger = logger

    def get_dist(self, p1, p2):
        return np.linalg.norm(p1 - p2, axis=-1)

    def get_tangent(self, guess, mask, cur, prev, next):

        cur_geom, prev_geom, next_geom = guess[:, cur], guess[:, prev], guess[:, next]

        prev_energy = self.image_potential(prev_geom, mask)
        cur_energy = self.image_potential(cur_geom, mask)
        next_energy = self.image_potential(next_geom, mask)
        if next_energy > cur_energy and cur_energy > prev_energy:
            tangent = next_geom - cur_geom
        elif next_energy <= cur_energy and cur_energy <= prev_energy:
            tangent = cur_geom - prev_geom
        else:
            dnext = abs(next_energy - cur_energy)
            dprev = abs(cur_energy - prev_energy)
            vmax = max(dnext, dprev) / (dnext + dprev)
            vmin = min(dnext, dprev) / (dnext + dprev)
            if next_energy > prev_energy:
                tangent = (next_geom - cur_geom) * vmax + (cur_geom - prev_geom) * vmin
            else:
                tangent = (next_geom - cur_geom) * vmin + (cur_geom - prev_geom) * vmax

        return vec_ops.vec_normalize(tangent)

    def adjust_jacobian(self, jac, guess, mask, cur, prev, next):
        if prev is None or next is None: return jac
        self._last_tangent = self.get_tangent(guess, mask, cur, prev, next)

        return vec_ops.project_out(jac, self._last_tangent[:, :, np.newaxis], orthonormal=True)

    def image_pairwise_contribution(self, guess, mask, cur, prev, next, order=0):
        if order > 2: return 0

        if prev is None or next is None:
            return 0

        cur_geom, prev_geom, next_geom = guess[:, cur], guess[:, prev], guess[:, next]

        prev_dist = self.get_dist(cur_geom, prev_geom)
        next_dist = self.get_dist(cur_geom, next_geom)

        dist = prev_dist - next_dist
        if misc.is_numeric(self.spring_constants):
            const = self.spring_constants
        else:
            const = self.spring_constants[cur]

        contribution = 0
        if order == 0:
            contribution = (const/2) * (dist**2)
        elif order == 1:
            tangent = self._last_tangent
            contribution = const * dist[..., np.newaxis] * tangent
        elif order == 2:
            contribution = const * vec_ops.identity_tensors(guess.shape[0], guess.shape[1])
        return contribution

class AdjustedChainStepFinder(ChainMinimizingStepFinder):
    def __init__(self,
                 pairwise_image_function,
                 func,
                 jacobian,
                 hessian=None,
                 logger=None,
                 **opts
                 ):
        super().__init__(
            func,
            jacobian,
            hessian=hessian,
            **opts
        )
        self.pairwise_function = pairwise_image_function
        self.logger = logger

    def image_pairwise_contribution(self,guess, mask, cur, prev, next, order=0):
        return self.pairwise_function(guess, mask, cur, prev, next, order=order)

class ChainReparametrizer:
    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> '(np.ndarray, np.ndarray, np.ndarray)':
        ...

class InterpolatingReparametrizer(ChainReparametrizer):
    #TODO: add in some kind of density function
    def __init__(self, interpolator_type):
        self.interpolator_type = interpolator_type

    def __call__(self, images, fixed_positions):
        interp = self.interpolator_type(images)
        new = interp(np.linspace(len(images)))
        interp[fixed_positions] = images[fixed_positions]
        return new


def jacobi_maximize(initial_matrix, rotation_generator, max_iterations=100, contrib_tol=1e-16, tol=1e-8):
    mat = np.asanyarray(initial_matrix).copy()

    k = initial_matrix.shape[1]
    perms = list(itertools.combinations(range(k), 2))
    U = np.eye(k)

    total_delta = -1
    iteration = -1
    for iteration in range(max_iterations):
        total_delta = 0
        for n, (p_i, q_i) in enumerate(perms):
            A, B, delta = rotation_generator(mat, p_i, q_i)

            if delta > 0:
                total_delta += delta

                new_pi = A * mat[:, p_i] + B * mat[:, q_i]
                new_qi = A * mat[:, q_i] - B * mat[:, p_i]
                mat[:, p_i] = new_pi
                mat[:, q_i] = new_qi

                rot_pi = A * U[:, p_i] + B * U[:, q_i]
                rot_qi = A * U[:, q_i] - B * U[:, p_i]
                U[:, p_i] = rot_pi
                U[:, q_i] = rot_qi

        if abs(total_delta) < tol:
            break

    return mat, U, (total_delta, iteration)

class LineSearchRotationGenerator:
    def __init__(self, column_function, tol=1e-16, max_iterations=10):
        self.one_e_func = column_function
        self.tol = tol
        self.max_iter = max_iterations

    @classmethod
    def quadratic_opt(self,
                  g0, g1, g2,
                  f0, f1, f2
                  ):
        g02 = g0**2
        g12 = g1**2
        g22 = g2**2
        denom = (2*f1*g0 - 2*f2*g0 - 2*f0*g1 + 2*f2*g1 + 2*f0*g2 - 2*f1*g2)
        if abs(denom) < 1e-8:
            return None
        else:
            return (
                    (f1*g02 - f2*g02 - f0*g12 + f2*g12 + f0*g22 - f1*g22)
                      / denom
            )

    def _phi(self, g, f_i, f_j):
        c = np.cos(g)
        s = np.sin(g)
        f_i, f_j = (
            c * f_i + s * f_j,
            -s * f_i + c * f_j
        )
        val = sum(self.one_e_func(f) for f in [f_i, f_j])
        return val, (c, s), (f_i, f_j)

    def __call__(self, mat, col_i, col_j):
        f_i, f_j = [mat[:, x] for x in [col_i, col_j]]
        phi0 = sum(self.one_e_func(f) for f in [f_i, f_j])

        g0 = 0
        g1 = np.pi
        g2 = 2*np.pi

        f0 = phi0
        f1, (c, s), _ = self._phi(g1, f_i, f_j)
        f2 = phi0

        prev = max([f0, f1, f2])
        for it in range(self.max_iter):
            g = self.quadratic_opt(
                  g0, g1, g2,
                  f0, f1, f2
                  )
            if g is None or g < g0 or g > g2:
                if f2 > f0:
                    g0 = g1
                    f0 = f1
                else:
                    g2 = g1
                    f2 = f1
                g = (g0 + g2) / 2
                f, (c, s), _ = self._phi(g, f_i, f_j)
            else:
                f, (c, s), _ = self._phi(g, f_i, f_j)
                if f <= min([f0, f1, f2]):
                    if f2 > f0:
                        g0 = g1
                        f0 = f1
                    else:
                        g2 = g1
                        f2 = f1
                    g = (g0 + g2) / 2
                    f, (c, s), _ = self._phi(g, f_i, f_j)
                else:
                    if g < g1:
                        g2 = g1
                        f2 = f1
                    else:
                        g0 = g1
                        f0 = f1
            # if abs(f - prev) < self.tol:
            #     break
            f1 = f
            g1 = g

        return c, s, f1 - prev

class GradientDescentRotationGenerator:
    def __init__(self, column_function, gradient, tol=1e-16, max_iterations=10,
                 damping_parameter=.9,
                 damping_exponent=1.1,
                 restart_interval=3
                 ):
        self.one_e_func = column_function
        self.grad = gradient
        self.tol = tol
        self.max_iter = max_iterations
        self.damper = Damper(
            damping_parameter=damping_parameter,
            damping_exponent=damping_exponent,
            restart_interval=restart_interval
        )

    def __call__(self, mat, col_i, col_j):
        f_i, f_j = [mat[:, x] for x in [col_i, col_j]]
        cur_val = sum(self.one_e_func(f) for f in [f_i, f_j])

        g = 0
        c = 1
        s = 0
        cur_grads = np.array([self.grad(f) for f in [f_i, f_j]])

        new_i, new_j = f_i, f_j
        for it in range(self.max_iter):
            grads = [np.dot(f, g) for f,g in zip([new_i, new_j], cur_grads)]
            step = sum(grads)
            if abs(step) < self.tol:
                break
            else:
                u = self.damper.get_damping_factor()
                if u is not None:
                    step *= u
                g = (g + step) % (2*np.pi)
                # if abs(g) > np.pi/2: g = np.sign(g) * np.pi/2
                c = np.cos(g)
                s = np.sin(g)
                new_i, new_j = (
                    c * f_i + s * f_j,
                    -s * f_i + c * f_j
                )
                cur_grads = np.array([self.grad(f) for f in [new_i, new_j]])


        new_vals = sum(self.one_e_func(f) for f in [new_i, new_j])
        return c, s, new_vals - cur_val

class OperatorMatrixRotationGenerator:
    def __init__(self, one_e_func, matrix_func):
        self.one_e_func = one_e_func
        self.mat_func = matrix_func
    def __call__(self, mat, col_i, col_j):
        f_i, f_j = [mat[:, x] for x in [col_i, col_j]]
        cur_val = sum(self.one_e_func(f) for f in [f_i, f_j])
        a, b, c = self.mat_func(f_i, f_j)

        test_mat = np.array([[a, b], [b, c]])
        # rot = np.linalg.eigh(test_mat)[1] # do this analytically...
        # print(rot)
        # cos_g = rot[0, 0]
        # sin_g = np.sign(rot[0, 0] * rot[1, 1]) * rot[1, 0]
        # new_rot = np.array([
        #     [cos_g, -sin_g],
        #     [sin_g, cos_g]
        # ])
        # explicit 2x2 form
        tau = (c - a) / (2 * b)
        t = np.sign(tau) / (abs(tau) + np.sqrt(1 + tau ** 2))
        cos_g = 1 / np.sqrt(1 + t ** 2)
        sin_g = -cos_g * t
        # new_rot = np.array([
        #         [cos_g, -sin_g],
        #         [sin_g, cos_g]
        #     ])
        # print(new_rot.T @ test_mat @ new_rot)

        f_i, f_j = (
            cos_g * f_i + sin_g * f_j,
            -sin_g * f_i + cos_g * f_j
        )
        new_val = sum(self.one_e_func(f) for f in [f_i, f_j])


        return cos_g, sin_g, new_val - cur_val

def displacement_localizing_rotation_generator(mat, col_i, col_j):
    # Foster-Boys localization

    p = mat[:, col_i].reshape(-1, 3)
    q = mat[:, col_j].reshape(-1, 3)
    pq_norms = vec_ops.vec_dots(p, q, axis=-1)
    pp_norms = vec_ops.vec_dots(p, p, axis=-1)
    qq_norms = vec_ops.vec_dots(q, q, axis=-1)

    pqpq = np.dot(pq_norms, pq_norms)
    pppp = np.dot(pp_norms, pp_norms)
    qqqq = np.dot(qq_norms, qq_norms)
    ppqq = np.dot(pp_norms, qq_norms)
    pppq = np.dot(pp_norms, pq_norms)
    qqqp = np.dot(qq_norms, pq_norms)

    A = pqpq - (pppp + qqqq - 2 * ppqq) / 4
    B = pppq - qqqp

    AB_norm = np.sqrt(A ** 2 + B ** 2)

    return A / AB_norm, B / AB_norm, A



import abc, os
from .. import Devutils as dev
from .Serializers import *

__all__ = [
    "Checkpointer",
    "CheckpointerKeyError",
    "DumpCheckpointer",
    "JSONCheckpointer",
    "NumPyCheckpointer",
    "HDF5Checkpointer",
    "DictCheckpointer",
    "NullCheckpointer"
]

class CheckpointerKeyError(KeyError):
    ...

class Checkpointer(metaclass=abc.ABCMeta):
    """
    General purpose base class that allows checkpointing to be done easily and cleanly.
    Intended to be a passable object that allows code to checkpoint easily.
    """

    default_extension=""
    def __init__(self, checkpoint_file,
                 allowed_keys=None,
                 omitted_keys=None
                 ):
        self.checkpoint_file = checkpoint_file
        self.allowed_keys = allowed_keys
        self.omitted_keys = omitted_keys
        self._came_open = not isinstance(checkpoint_file, str)
        self._open_depth = 0
        self._stream = None
    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.checkpoint_file)

    _ext_map = None
    @classmethod
    def extension_map(cls):
        if cls._ext_map is not None:
            return cls._ext_map
        else:
            return {c.default_extension:c for c in [JSONCheckpointer, HDF5Checkpointer, NumPyCheckpointer]}

    @classmethod
    def build_canonical(cls, checkpoint):
        """
        Dispatches over types of objects to make a canonical checkpointer
        from the supplied data

        :param checkpoint: provides
        :type checkpoint: None | str | Checkpoint | file | dict
        :return:
        :rtype: Checkpointer
        """

        if checkpoint is None:
            return NullCheckpointer(None)
        elif isinstance(checkpoint, str):
            return Checkpointer.from_file(checkpoint)
        elif isinstance(checkpoint, Checkpointer):
            return checkpoint
        elif dev.is_dict_like(checkpoint):
            if dev.Schema(["file"]).validate(checkpoint, throw=False):
                checkpoint = dev.Schema(["file"], ['keys', 'opts']).to_dict(checkpoint)
                opts = checkpoint['opts'] if 'opts' in checkpoint else {}
                if 'keys' in checkpoint:
                    allowed_opts = dev.Schema(['allowed_keys'], ['omitted_keys']).to_dict(checkpoint['keys'], throw=False)
                    if allowed_opts is not None:
                        opts = dict(opts, **allowed_opts)
                    else:
                        omitted_opts = dev.Schema(['omitted_keys']).to_dict(checkpoint['keys'], throw=False)
                        if allowed_opts is not None:
                            opts = dict(opts, **omitted_opts)
                        else:
                            opts['allowed_keys'] = checkpoint['keys']
                return cls.from_file(checkpoint['file'], **opts)
            else:
                return DictCheckpointer(checkpoint)
        else:
            return checkpoint

    @classmethod
    def from_file(cls, file, **opts):
        """
        Dispatch function to load from the appropriate file
        :param file:
        :type file: str | File
        :param opts:
        :type opts:
        :return:
        :rtype:
        """

        if not isinstance(file, str):
            #TODO: make this cleaner
            file_name = file.name # might break in the future...
        else:
            file_name = file

        _, ext = os.path.splitext(file_name)

        ext_map = cls.extension_map()

        if ext not in ext_map:
            raise ValueError("don't know have default checkpointer type registered for extension {}".format(ext))

        return ext_map[ext](file, **opts)

    def __enter__(self):
        self._open_depth+=1
        if self._stream is None:
            self._stream = self.open_checkpoint_file(self.checkpoint_file)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._open_depth-=1
        if self._stream is not None:
            if self._open_depth == 0:
                self.close_checkpoint_file(self._stream)
                self._stream = None

    def cached_eval(self, key, generator, *,
                    condition=None,
                    args=(),
                    kwargs=None):
        return dev.cached_eval(
            self,
            key,
            generator,
            condition=condition,
            args=args,
            kwargs=kwargs
        )

    @property
    def is_open(self):
        return self._stream is not None

    @property
    def stream(self):
        return self._stream

    @abc.abstractmethod
    def open_checkpoint_file(self, chk):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk:
        :return:
        :rtype:
        """
        raise NotImplementedError("CheckpointerBase is an abstract base class...")
    @abc.abstractmethod
    def close_checkpoint_file(self, stream):
        """
        Closes the opened checkpointing stream
        :param stream:
        :type stream:
        :return:
        :rtype:
        """
        raise NotImplementedError("CheckpointerBase is an abstract base class...")
    @abc.abstractmethod
    def save_parameter(self, key, value):
        """
        Saves a parameter to the checkpoint file
        :param key:
        :type key:
        :param value:
        :type value:
        :return:
        :rtype:
        """
        raise NotImplementedError("CheckpointerBase is an abstract base class...")
    @abc.abstractmethod
    def load_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        raise NotImplementedError("CheckpointerBase is an abstract base class...")

    def delete_parameter(self, key):
        raise NotImplementedError("checkpointer does not need to implement deletion")
    def check_parameter(self, key):
        try:
            self.check_allowed_key(key)
            self.__getitem__(key)
        except KeyError:
            return False
        else:
            return True

    def update(self, vals):
        if dev.is_dict_like(vals):
            return self.update(vals.items())

        if not self.is_open:
            with self:
                return self.update(vals)

        for k, v in vals:
            self[k] = v
    def get_keys(self, keys):
        if not self.is_open:
            with self:
                return self.get_keys(keys)

        return [
            self[k] for k in keys
        ]

    def check_allowed_key(self, item):
        if isinstance(item, tuple):
            # for subkey in item:
            self.check_allowed_key(item[0]) # only check topline items
        else:
            if self.allowed_keys is not None:
                if item not in self.allowed_keys:
                    raise CheckpointerKeyError("key {} not allowed by {}".format(
                        item,
                        self
                    ))
            if self.omitted_keys is not None:
                if item in self.omitted_keys:
                    raise CheckpointerKeyError("key {} not allowed by {}".format(
                        item,
                        self
                    ))

    def __contains__(self, key):
        if not self.is_open:
            with self:
                return self.__contains__(key)
        return self.check_parameter(key)

    def __getitem__(self, item):
        if not self.is_open:
            with self:
                return self.__getitem__(item)
        self.check_allowed_key(item)
        return self.load_parameter(item)
    def get(self, key, default=None):
        try:
            val = self[key]
        except KeyError:
            val = default
        return val
    def __setitem__(self, key, value):
        if not self.is_open:
            with self:
                return self.__setitem__(key, value)
        self.check_allowed_key(key)
        self.save_parameter(key, value)

    def __delitem__(self, key):
        if not self.is_open:
            with self:
                return self.__delitem__(key)
        self.check_allowed_key(key)
        self.delete_parameter(key)
    def pop(self, key, *default):
        if len(default) > 0:
            try:
                val = self[key]
            except KeyError:
                val = default[0]
            else:
                del self[key]
        else:
            val = self[key]
            del self[key]
        return val

    @abc.abstractmethod
    def keys(self):
        """
        Returns the keys of currently checkpointed
        objects

        :return:
        :rtype:
        """
        raise NotImplementedError("Checkpointer is an abstract base class...")

class DumpCheckpointer(Checkpointer):
    """
    A subclass of `CheckpointerBase` that writes an entire dump to file at once & maintains
    a backend cache to update it cleanly
    """
    def __init__(self, file, cache=None, open_kwargs=None,
                 allowed_keys=None,
                 omitted_keys=None
                 ):
        self.backend = cache # cache values
        super().__init__(file, allowed_keys=allowed_keys, omitted_keys=omitted_keys)
        if open_kwargs is None:
            open_kwargs = {'mode':"w+"}
        self.open_kwargs = open_kwargs
    def load_cache(self):
        if self.backend is None:
            self.backend = {}
    def __enter__(self):
        self.load_cache()
        return super().__enter__()
    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.dump()
        finally:
            super().__exit__(exc_type, exc_val, exc_tb)
    @abc.abstractmethod
    def dump(self):
        """
        Writes the entire data structure
        :return:
        :rtype:
        """
        raise NotImplementedError("DumpCheckpointer is an ABC and doesn't know how to write to file")
    def convert(self):
        """
        Converts the cache to an exportable form if needed
        :return:
        :rtype:
        """
        return self.backend
    def open_checkpoint_file(self, chk):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk:
        :return:
        :rtype:
        """
        if isinstance(chk, str):
            chk = open(chk, **self.open_kwargs)
        return chk
    def close_checkpoint_file(self, stream):
        """
        Closes the opened checkpointing stream
        :param stream:
        :type stream:
        :return:
        :rtype:
        """
        if not self._came_open:
            stream.close()
    def save_parameter(self, key, value):
        """
        Saves a parameter to the checkpoint file
        :param key:
        :type key:
        :param value:
        :type value:
        :return:
        :rtype:
        """
        if isinstance(key, tuple):
            base = self.backend
            cur = key[0]
            for subk in key[1:]:
                if cur not in base: base[cur] = {}
                base = base[cur]
                cur = subk
            base[cur] = value
        else:
            self.backend[key] = value
    def check_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        if isinstance(key, tuple):
            base = self.backend
            cur = key[0]
            for subk in key[1:]:
                if cur not in base: return False
                base = base[cur]
                cur = subk
            return cur in base
        else:
            return key in self.backend
    def load_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        if isinstance(key, tuple):
            base = self.backend
            cur = key[0]
            for subk in key[1:]:
                if cur not in base:
                    raise KeyError(f"{key} not found")
                base = base[cur]
                cur = subk
            val = base[cur]
        else:
            val = self.backend[key]
        return val
    def delete_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        if isinstance(key, tuple):
            base = self.backend
            cur = key[0]
            for subk in key[1:]:
                if cur not in base:
                    raise KeyError(f"{key} not found")
                base = base[cur]
                cur = subk
            del base[cur]
        else:
            del self.backend[key]

    def keys(self):
        if not self.is_open:
            with self:
                return self.keys()
        return self.backend.keys()

class JSONCheckpointer(DumpCheckpointer):
    """
    A checkpointer that uses JSON as a backend
    """

    default_extension=JSONSerializer.default_extension
    def __init__(self, file, cache=None, serializer=None, open_kwargs=None,
                 allowed_keys=None,
                 omitted_keys=None
                 ):
        if serializer is None:
            serializer = JSONSerializer()
        self.serializer = serializer
        super().__init__(file, cache=cache, open_kwargs=open_kwargs, allowed_keys=allowed_keys, omitted_keys=omitted_keys)

    def load_cache(self):
        cache = self.backend
        if cache is None:
            file = self.checkpoint_file
            serializer = self.serializer
            if isinstance(file, str) and os.path.exists(file) and os.stat(file).st_size > 0:
                with open(file, 'r') as stream:
                    cache = serializer.deserialize(stream)
                if not isinstance(cache, dict):
                    cache = {}
            else:
                cache = {}
            self.backend = cache

    def dump(self):
        """
        Writes the entire data structure
        :return:
        :rtype:
        """
        self.serializer.serialize(self.stream, self.backend)

class NumPyCheckpointer(DumpCheckpointer):
    """
    A checkpointer that uses NumPy as a backend
    """

    default_extension = NumPySerializer.default_extension
    def __init__(self, file, cache=None, serializer=None, open_kwargs=None,
                 allowed_keys=None,
                 omitted_keys=None
                 ):
        if isinstance(file, str):
            if not os.path.exists(file):
                if os.path.exists(file + '.npz'):
                    file = file + '.npz'
                elif os.path.exists(file + '.npy'):
                    file = file + '.npy'

        if serializer is None:
            serializer = NumPySerializer()
        self.serializer = serializer
        if open_kwargs is None:
            open_kwargs = {'mode':'bw'}
        super().__init__(file, cache=cache, open_kwargs=open_kwargs,
                         allowed_keys=allowed_keys,
                         omitted_keys=omitted_keys
                         )

    def load_cache(self):
        cache = self.backend
        if cache is None:
            file = self.checkpoint_file
            serializer = self.serializer
            if isinstance(file, str) and os.path.exists(file) and os.stat(file).st_size > 0:
                with open(file, 'br') as stream:
                    cache = serializer.deserialize(stream)
                if not isinstance(cache, dict):
                    cache = {}
            else:
                cache = {}
            self.backend = cache

    def dump(self):
        """
        Writes the entire data structure
        :return:
        :rtype:
        """
        self.serializer.serialize(self.stream, self.backend)

class HDF5Checkpointer(Checkpointer):
    """
    A checkpointer that uses an HDF5 file as a backend.
    Doesn't maintain a secondary `dict`, because HDF5 is an updatable format.
    """

    default_extension = HDF5Serializer.default_extension
    def __init__(self, checkpoint_file, serializer=None,
                 allowed_keys=None,
                 omitted_keys=None
                 ):
        super().__init__(checkpoint_file, allowed_keys=allowed_keys, omitted_keys=omitted_keys)
        if serializer is None:
            serializer = HDF5Serializer()
        self.serializer = serializer

    def open_checkpoint_file(self, chk):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk: str | file-like
        :return:
        :rtype:
        """
        if not self._came_open:
            try:
                if os.path.exists(chk):
                    return open(chk, 'r+b')
                else:
                    return open(chk, "w+b")
            except ValueError as e:
                if e.args[0] == 'seek of closed file':
                    raise IOError("existing HDF5 file {} is corrupted and can't be opened".format(chk))
                else:
                    raise
        elif 'b' not in chk.mode:
            raise IOError("{} isn't opened in binary mode (HDF5 needs that)".format(chk))
        else:
            return chk

    def close_checkpoint_file(self, stream):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk:
        :return:
        :rtype:
        """
        if not self._came_open:
            self.stream.close()

    def save_parameter(self, key, value):
        """
        Saves a parameter to the checkpoint file
        :param key:
        :type key:
        :param value:
        :type value:
        :return:
        :rtype:
        """
        # HDF5 serialization is an updateable process
        if self.stream is None:
            raise IOError("stream for {} got closed and won't reopen".format(self.checkpoint_file))
        self.serializer.serialize(self.stream, {key:value})

    def load_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        return self.serializer.deserialize(self.stream, key=key)

    def keys(self):
        if not self.is_open:
            with self:
                return self.keys()
        file = self.stream
        if not isinstance(file, (self.serializer.api.File, self.serializer.api.Group)):
            file = self.serializer.api.File(file, "a")
        return list(file.keys())

class DictCheckpointer(Checkpointer):
    """
    A checkpointer that doesn't actually do anything, but which is provided
    so that programs can turn off checkpointing without changing their layout
    """
    def __init__(self, checkpoint_file=None,
                 allowed_keys=None,
                 omitted_keys=None
                 ):
        if checkpoint_file is None:
            checkpoint_file = {}
        super().__init__(checkpoint_file, allowed_keys=allowed_keys, omitted_keys=omitted_keys)
        self.backend = checkpoint_file

    def open_checkpoint_file(self, chk):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk: str | file-like
        :return:
        :rtype:
        """
        return chk

    def close_checkpoint_file(self, stream):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk:
        :return:
        :rtype:
        """
        pass

    def save_parameter(self, key, value):
        """
        Saves a parameter to the checkpoint file
        :param key:
        :type key:
        :param value:
        :type value:
        :return:
        :rtype:
        """
        self.backend[key] = value

    def load_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        return self.backend[key]
    def delete_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        del self.backend[key]

    def keys(self):
        return list(self.backend.keys())
    def get(self, key, default=None):
        return self.backend.get(key, default)
    def pop(self, key, *default):
        return self.backend.pop(key, *default)

class NullCheckpointer(Checkpointer):
    """
    A checkpointer that saves absolutely nothing
    """
    def __init__(self, checkpoint_file=None,
                 allowed_keys=None,
                 omitted_keys=None
                 ):
        super().__init__(checkpoint_file, allowed_keys=allowed_keys, omitted_keys=omitted_keys)
        self.backend = None

    def open_checkpoint_file(self, chk):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk: str | file-like
        :return:
        :rtype:
        """
        return "NotAFile"

    def close_checkpoint_file(self, stream):
        """
        Opens the passed `checkpoint_file` (if not already open)
        :param chk:
        :type chk:
        :return:
        :rtype:
        """
        pass

    def save_parameter(self, key, value):
        """
        Saves a parameter to the checkpoint file
        :param key:
        :type key:
        :param value:
        :type value:
        :return:
        :rtype:
        """
        pass

    def load_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        raise CheckpointerKeyError("NullCheckpointer doesn't support _any_ keys")

    def delete_parameter(self, key):
        """
        Loads a parameter from the checkpoint file
        :param key:
        :type key:
        :return:
        :rtype:
        """
        raise CheckpointerKeyError("NullCheckpointer doesn't support _any_ keys")

    def keys(self):
        return []



import importlib, importlib.abc, os, importlib.util, sys

__all__ = [
    "ModuleLoader"
]

class DynamicModuleLoader(importlib.abc.SourceLoader):
    """
    A `DynamicModuleLoader` is a Loader object that can load a python module from a file path.
    Primarily intended for workflows that don't support `importlib.import_module`
    """
    tag = "DynamicImports"
    def __init__(self, rootdir='', rootpkg=None, retag=True):
        """
        :param rootdir: root directory to look for files off of
        :type rootdir: str
        :param rootpkg: root package to look for files off of
        :type rootpkg: str or None
        """
        self._dir=rootdir
        self._pkg = rootpkg
        if retag is True:
            retag = self.tag
        self._reg=retag
        super().__init__()

    def get_data(self, file):
        with open(file,'rb') as src:
            return src.read()

    def get_filename(self, fullname):
        if not os.path.exists(fullname):
            basename, ext = os.path.splitext(fullname.split(".")[-1])
            if ext != "":
                fullname = os.path.join(self._dir, basename+ext)
            else:
                fullname = os.path.join(self._dir, basename+".py")
                if not os.path.exists(fullname):
                    fullname = os.path.join(self._dir, basename)
        if os.path.isdir(fullname):
            fullname = os.path.join(fullname, "__init__.py")
        return fullname

    def get_spec(self, file, pkg = None):
        base_name = os.path.splitext(os.path.basename(file))[0]
        package_origin = file
        if pkg is None:
            pkg = self._pkg
        if pkg is None:
            raise ImportError("{}: package name required to load file".format(type(self)))
        if pkg == "":
            package_name = base_name
        else:
            package_name = pkg + "." + base_name
        spec = importlib.util.spec_from_loader(
            package_name,
            self,
            origin=package_origin
        )
        return spec

    def reregister_module(self, module, tag=None):
        """
        Sets up a secondary hook for a module so it's clear which
        ones were dynamically loaded

        :param tag:
        :type tag:
        :return:
        :rtype:
        """
        if tag is None:
            tag = self._reg
        sys.modules[tag + "." + module.__name__] = module

    def load(self, file, pkg=None):
        """
        loads a file as a module with optional package name

        :param file:
        :type file: str
        :param pkg:
        :type pkg: str or None
        :return:
        :rtype: module
        """
        d = self._dir
        try:
            if os.path.exists(file):
                self._dir = os.path.dirname(file)
            # print(file, self._dir)
            spec = self.get_spec(file, pkg)
            module = importlib.util.module_from_spec(spec)
            if module is None:
                module = importlib.util.module_from_spec(None)
            self.exec_module(module)
            self.reregister_module(module)
        finally:
            self._dir = d
        return module

class ModuleLoader:
    """
    Provides a way to load dynamic modules.
    Either use a `DynamicModuleLoader` or the `importlib.import_module` function
    depending on how much customization is needed.
    """
    def __init__(self, rootdir='', rootpkg=None, retag=False):
        """
        :param rootdir: root directory to look for files off of
        :type rootdir: str
        :param rootpkg: root package to look for files off of
        :type rootpkg: str or None
        """
        # backwards compatibility
        if not hasattr(importlib, 'import_module') and not retag:
            retag = True
        if retag:
            self.loader = DynamicModuleLoader(rootdir=rootdir, rootpkg=rootpkg, retag=retag)
        else:
            self.root_dir = rootdir
            self.loader = None

    def load(self, file, pkg=None):
        if self.loader is not None:
            return self.loader.load(file, pkg=pkg)
        else:
            rootdir = os.path.dirname(file)
            if rootdir == "" or not os.path.isdir(rootdir):
                rootdir = self.root_dir
            if rootdir is not None:
                sys.path.insert(0, rootdir)
            try:
                if pkg is None:
                    file = os.path.splitext(file)[0]
                    mod = importlib.import_module(file)
                else:
                    mod = importlib.import_module(file, package=pkg)
            finally:
                if rootdir is not None:
                    sys.path.pop(0)
            return mod



"""
Provides some support for working with the python bindings for external programs, like OpenBabel
Mostly relevant for doing format conversions/parsing, but other utilities do exist.
"""

__all__ = []
from .Jobs import *; from .Jobs import __all__ as exposed
__all__ += exposed
from .Parsers import *; from .Parsers import __all__ as exposed
__all__ += exposed
from .Runner import *; from .Runner import __all__ as exposed
__all__ += exposed
from .ChemToolkits import *; from .ChemToolkits import __all__ as exposed
__all__ += exposed
from .ImageKits import *; from .ImageKits import __all__ as exposed
__all__ += exposed
from .Toolkits3D import *; from .Toolkits3D import __all__ as exposed
__all__ += exposed
from .Visualizers import *; from .Visualizers import __all__ as exposed
__all__ += exposed
from .RDKit import *; from .RDKit import __all__ as exposed
__all__ += exposed
from .ASE import *; from .ASE import __all__ as exposed
__all__ += exposed
from .OpenBabel import *; from .OpenBabel import __all__ as exposed
__all__ += exposed
from .WebAPI import *; from .WebAPI import __all__ as exposed
__all__ += exposed
from .ChemSpiderAPI import *; from .ChemSpiderAPI import __all__ as exposed
__all__ += exposed
from .Subprocesses import *; from .Subprocesses import __all__ as exposed
__all__ += exposed
from .Servers import *; from .Servers import __all__ as exposed
__all__ += exposed


import uuid

import numpy as np
import urllib.parse
from ..Data import ColorData
from .. import Numputils as nput
from .. import Devutils as dev

__all__ = [
    "ColorPalette"
]

#TODO: add ColorSpaces enum for validation

class ColorPalette:
    def __init__(self, colors, blend_spacings=None, lab_colors=None, color_space='rgb', cycle=False):
        if isinstance(colors, str):
            colors = self.resolve_color_palette(colors)
        elif self.is_colormap_like(colors):
            colors = self.discretize_colormap(colors)
        if not self.is_palette_list(colors):
            raise ValueError(f"{colors} is not a color palette list")
        self.color_strings, self.lab_colors = self.prep_color_palette(colors, color_space, lab_colors=lab_colors)
        # TODO: add more sophisticated blending
        if blend_spacings is None:
            blend_spacings = np.linspace(0, 1, len(self.color_strings))
        self.abcissae = np.asanyarray(blend_spacings)
        self.cycle = cycle

    def __hash__(self):
        return hash((type(self), self.color_strings))

    named_colors = {}
    @classmethod
    def parse_color_string(cls, name:str):
        if not name.startswith('#'):
            c = cls.named_colors.get(name)
            if c is None:
                from matplotlib.colors import to_rgba
                c = [255*x for x in to_rgba(name)[:3]]
        else:
            c = name
        if isinstance(c, str):
            c = cls.parse_rgb_code(c)
        return c

    @classmethod
    def prep_color_palette(cls, colors, color_space='rgb', lab_colors=None):
        if lab_colors is not None:
            lab_colors = np.asanyarray(lab_colors)
        if isinstance(colors[0], str):
            if color_space != 'rgb':
                raise ValueError(f"only rgb color codes supported (got {colors})")
            color_lists = colors
            rgb_array = np.array([cls.parse_rgb_code(c) for c in colors])
            if lab_colors is None:
                lab_colors = cls.color_convert(rgb_array.T, "rgb", "lab").T
        else:
            colors = np.asanyarray(colors)
            if lab_colors is None:
                if color_space != 'lab':
                    lab_colors = cls.color_convert(colors.T, color_space, "lab").T
                else:
                    lab_colors = colors
            if color_space != 'rgb':
                rgb_colors = cls.color_convert(colors.T, color_space, 'rgb').T
            else:
                rgb_colors = colors
            color_lists = [cls.rgb_code(c, 2) for c in rgb_colors]

        return tuple(color_lists), lab_colors

    @classmethod
    def resolve_color_palette(cls, cmap_name):
        try:
            data = ColorData[cmap_name].data
        except KeyError:
            from matplotlib import colormaps

            data = cls.discretize_colormap(colormaps[cmap_name])

        return data

    @classmethod
    def is_colormap_like(cls, cmap):
        return hasattr(cmap, '__call__')
    @classmethod
    def discretize_colormap(cls, cmap, samples=10):
        if isinstance(cmap, ColorPalette):
            return cmap.color_strings
        else:
            vals = cmap(np.linspace(0, 1, samples))
            return 255 * vals[:, :3]

    @classmethod
    def is_palette_list(self, colors):
        return (
                dev.is_list_like(colors)
                and all(
                    isinstance(c, str)
                    or nput.is_numeric_array_like(c)
                    for c in colors
                )
        )
    def flip(self):
        return type(self)(list(reversed(self.color_strings)))

    def __eq__(self, other):
        if not hasattr(other, "color_strings"): return False
        return self.color_strings == other.color_strings

    def get_colorblindness_test_url(self):
        return "https://davidmathlogic.com/colorblind/#" + "-".join(
            urllib.parse.quote(c[:7] if c.startswith("#") else c[:6]) for c in self.color_strings
        )

    def blend(self, amount, modification_space='lab', rescale=False, return_color_code=True):
        amount = np.asanyarray(amount)
        smol = amount.ndim == 0
        if smol: amount = np.array([amount])
        insertion_indices = np.searchsorted(self.abcissae, amount)
        terminals = insertion_indices == len(self.abcissae)
        starts = insertion_indices == 0
        rems = np.logical_not(np.logical_or(terminals, starts))
        codes = [""] * len(amount)
        new_colors = np.empty((len(amount), len(self.lab_colors[0])), dtype=float)
        if return_color_code:
            term_pos = np.where(terminals)
            if len(term_pos) > 0:
                for i in term_pos[0]:
                    codes[i] = self.color_strings[-1]
            start_pos = np.where(starts)
            if len(start_pos) > 0:
                for i in start_pos[0]:
                    codes[i] = self.color_strings[0]
        else:
            if terminals.any():
                color = np.array(self.parse_rgb_code(self.color_strings[-1]))
                new_colors[terminals] = color[np.newaxis]
            if starts.any():
                color = np.array(self.parse_rgb_code(self.color_strings[0]))
                new_colors[starts] = color[np.newaxis]
        if rems.any():
            rem_inds = insertion_indices[rems]
            x = self.abcissae[rem_inds - 1,]
            y = self.abcissae[rem_inds,]
            d = ((amount[rems,] - x) / (y - x))[:, np.newaxis]

            colors = self.lab_colors
            if modification_space != 'lab':
                colors = self.color_convert(colors, 'lab', modification_space).T
            new_lab = colors[rem_inds - 1,] * (1 - d) + colors[rem_inds,] * d
            rgb = self.color_convert(new_lab.T, modification_space, 'rgb').T

            if return_color_code:
                rgb = self.rgb_code(rgb.T)
                rems = np.where(rems)[0]
                for n,r in enumerate(rgb):
                    codes[rems[n]] = r
            else:
                new_colors[rems,] = rgb

        if return_color_code:
            if smol: codes = codes[0]
            return codes
        else:
            if rescale: new_colors = self.color_rescale(new_colors, 'rgb')
            if smol: new_colors = new_colors[0]
            return new_colors

    def as_colormap(self, levels=None, cmap_type='list', name=None, **opts):
        from matplotlib.colors import ListedColormap, LinearSegmentedColormap

        if levels is None:
            levels = np.linspace(0, 1, len(self.color_strings))
            vals = np.array(self.parse_rgb_code(self.color_strings)) / 255
        else:
            if nput.is_int(levels):
                levels = np.linspace(0, 1, levels)
            vals = self(levels)

        if dev.str_is(cmap_type, 'list'):
            cmap = np.concatenate([
                vals,
                np.ones((len(vals), 1)),
                ],
                axis=1
            )
            new_map = ListedColormap(cmap, **opts)
        elif dev.str_is(cmap_type, 'interpolated'):
            cmap_dict = {
                'red':np.concatenate([
                    levels[:, np.newaxis],
                    vals[:, (0,)],
                    vals[:, (0,)]
                ],axis=1),
                'green':np.concatenate([
                    levels[:, np.newaxis],
                    vals[:, (1,)],
                    vals[:, (1,)]
                ],axis=1),
                'blue':np.concatenate([
                    levels[:, np.newaxis],
                    vals[:, (2,)],
                    vals[:, (2,)]
                ],axis=1)
            }
            if name is None:
                name = '-'.join(['cmap']+str(uuid.uuid4()).split("-")[:2])
            new_map = LinearSegmentedColormap(name, segmentdata=cmap_dict, **opts)
        else:
            new_map: 'ListedColormap|LinearSegmentedColormap' = cmap_type(levels, vals, **opts)

        return new_map

    def __call__(self, amount, rescale=True, return_color_code=False):
        return self.blend(amount, rescale=rescale, return_color_code=return_color_code)

    def modify(self, modification_function, modification_space='lab', clip=True):
        return type(self)(
            modification_function(
                self.lab_colors.T,
                color_space='lab',
                modification_space=modification_space,
                clip=clip
            ).T,
            color_space='lab'
        )

    def lighten(self, percentage,
                modification_space='lab',
                shift=False,
                absolute=False, clip=True):
        return type(self)(
            self.color_lighten(self.lab_colors.T,
                               percentage,
                               color_space='lab',
                               modification_space=modification_space,
                               shift=shift,
                               absolute=absolute,
                               clip=clip
                               ).T,
            color_space='lab'
        )

    @classmethod
    def color_normalize(cls, color_list, color_space='rgb'):
        color_list = np.asanyarray(color_list)
        smol = color_list.ndim == 1
        if smol: color_list = color_list[:, np.newaxis]
        if color_space == 'rgb':
            color_list = np.clip(color_list, 0, 255)
        # elif color_space == 'xyz':
        #     color_list = np.clip(color_list, 0, 100)
        elif color_space in {'hsl', 'hsv'}:
            color_list = np.clip(color_list, 0, 1)
        if smol: color_list = color_list[:, 0]
        return color_list

    @classmethod
    def color_rescale(cls, color_list, color_space='rgb'):
        color_list = np.asanyarray(color_list)
        if color_space == 'rgb':
            color_list = color_list / 255
        elif color_space == 'xyz':
            color_list = color_list / 100
        return color_list
    @classmethod
    def color_modify(cls, color, modification_function, color_space='rgb', modification_space='lab', clip=True):
        as_code = isinstance(color, str)
        padding = 2
        if as_code:
            if color_space != 'rgb':
                raise ValueError(f"only rgb color codes supported (got {color})")
            color, padding = cls.parse_rgb_code(color, return_padding=True)
        if color_space != modification_space:
            lab_color = cls.color_convert(color, color_space, modification_space)
        else:
            lab_color = color
        lab_color = modification_function(*lab_color)
        if color_space != modification_space:
            color = cls.color_convert(lab_color, modification_space, color_space)
        if clip:
            color = cls.color_normalize(color, color_space)

        if as_code:
            color = cls.rgb_code(color, padding=padding)
        return color
    @classmethod
    def color_lighten(cls, color, percentage,
                      color_space='rgb',
                      modification_space='lab',
                      shift=False,
                      absolute=False, clip=True):
        if modification_space == 'lab':
            if shift:
                percentage = 100*percentage
                conversion = lambda l,a,b:[l+percentage, a, b]
            elif absolute:
                percentage = 100*percentage
                conversion = lambda l,a,b:[percentage, a, b]
            else:
                conversion = lambda l,a,b:[l*(1+percentage), a, b]
        elif modification_space in {'hsv', 'hsl'}:
            if shift:
                conversion = lambda h,s,l:[h, s, l+percentage]
            elif absolute:
                conversion = lambda h,s,l:[h, s, percentage]
            else:
                conversion = lambda h,s,l:[h, s, l*(1+percentage)]
        else:
            raise ValueError(f"can't lighten color in modification_space `{modification_space}`")

        return cls.color_modify(color, conversion, color_space=color_space, modification_space=modification_space, clip=clip)

    def __len__(self):
        return len(self.color_strings)
    def __getitem__(self, item):
        if nput.is_int(item):
            if not self.cycle:
                return self.color_strings[item]
            else:
                return self.color_strings[item % len(self.color_strings)]
        else:
            return type(self)(
                self.color_strings[item],
                blend_spacings=self.abcissae[item],
                lab_colors=self.lab_colors[item]
            )


    @classmethod
    def rgb_code(cls, rgb, padding=2):
        if not isinstance(rgb[0], (int, float, np.floating, np.integer)):
            return [
                cls.rgb_code([r, g, b])
                for r, g, b in zip(*rgb)
            ]
        rgb = np.round(np.clip(rgb, 0, 255)).astype(int)
        return f"#{rgb[0]:0>{padding}x}{rgb[1]:0>{padding}x}{rgb[2]:0>{padding}x}"
    @classmethod
    def parse_rgb_code(cls, code, padding=None, return_padding=False):
        if not isinstance(code, str):
            if not return_padding:
                return [
                    cls.parse_rgb_code(c, padding=padding, return_padding=False)
                    for c in code
                ]
            else:
                padding, _ = cls.parse_rgb_code(code[0], padding=padding, return_padding=True)
                return [
                    cls.parse_rgb_code(c, padding=padding, return_padding=False)
                    for c in code
                ], padding
        if code[0] == "#":
            code = code[1:]
        if padding is None:
            padding = len(code) // 3
        color_list = [
            int(code[(padding*i):(padding*(i+1))], 16)
            for i in range(3)
        ]
        if return_padding:
            return color_list, padding
        else:
            return color_list

    converters = {}
    @classmethod
    def color_convert(self, color, original_space, target_space):
        if original_space == target_space:
            return color
        if (original_space, target_space) in self.converters:
            conversion = self.converters[(original_space, target_space)]
        else:
            if original_space == 'rgb' or target_space == 'rgb':
                conversion = getattr(self, f"{original_space}_to_{target_space}")
            else:
                try:
                    conversion = getattr(self, f"{original_space}_to_{target_space}") #TODO: register these better
                except AttributeError:
                    # send everything through RGB
                    conversion1 = getattr(self, f"{original_space}_to_rgb")
                    conversion2 = getattr(self, f"rgb_to_{target_space}")
                    conversion = lambda *c: conversion2(*conversion1(*c))
        return conversion(*color)

    xyz_to_rbg_array = [
        # exact-ish inverse of conversion matrix
        [
            670962301703 * (1000000 / 207056369298614928),
            -318277012021 * (1000000 / 207056369298614928),
            -103225121660 * (1000000 / 207056369298614928)
        ],
        [
            -200690410871 * (1000000 / 207056369298614928),
            388435678549 * (1000000 / 207056369298614928),
            8604419276 * (1000000 / 207056369298614928)
        ],
        [
            11521991063 * (1000000 / 207056369298614928),
            -42248058709 * (1000000 / 207056369298614928),
            218922991300 * (1000000 / 207056369298614928)
        ]
    ]
    @classmethod
    def xyz_to_rgb(self, x, y, z):
        # converted from https://www.easyrgb.com/en/math.php
        if not isinstance(self.xyz_to_rbg_array, np.ndarray):
            self.xyz_to_rbg_array = np.array(self.xyz_to_rbg_array)

        xyz = np.array([x, y, z]) / 100
        rgb = np.tensordot(self.xyz_to_rbg_array, xyz, axes=[0, 0])
        mask = rgb > 0.0031308
        rgb[mask] = 1.055*rgb[mask]**(1/2.4) - 0.055
        not_mask = np.logical_not(mask)
        rgb[not_mask] = rgb[not_mask] * 12.92
        return rgb * 255

    rgb_to_xyz_array = [ # just the inverse
        [0.412453, 0.357580, 0.180423],
        [0.212671, 0.715160, 0.072169],
        [0.019334, 0.119193, 0.950227],
    ]
    @classmethod
    def rgb_to_xyz(self, r, g, b):
        if not isinstance(self.rgb_to_xyz_array, np.ndarray):
            self.rgb_to_xyz_array = np.array(self.rgb_to_xyz_array)

        rgb = np.array([r, g, b]) / 255
        mask = rgb > 0.04045
        rgb[mask] = ((rgb[mask] + 0.055) / 1.055)**(2.4)
        not_mask = np.logical_not(mask)
        rgb[not_mask] = rgb[not_mask] / 12.92

        xyz = np.tensordot(self.rgb_to_xyz_array, rgb, axes=[0, 0])
        return xyz * 100

    @classmethod
    def _rgb2hue(cls, rgb, diff, max_val):

        diff_r, diff_g, diff_b = [
            ((max_val - c) / 6 + (diff / 2)) / diff
            for c in rgb
        ]
        r_primary, g_primary, b_primary = [
            max_val == c
            for c in rgb
        ]

        h = np.zeros_like(diff)
        h[r_primary] = (diff_b[r_primary,] - diff_g[r_primary,])
        h[g_primary] = (1 / 3 + diff_r[g_primary,] - diff_b[g_primary,])
        h[b_primary] = (2 / 3 + diff_g[b_primary,] - diff_r[b_primary,])

        h[h < 0] += 1
        h[h > 1] -= 1

        return h
    @classmethod
    def rgb_to_hsl(self, r, g, b):

        rgb = np.array([r, g, b]) / 255
        smol = rgb.ndim == 1
        if smol:
            rgb = rgb[:, np.newaxis]
        base_shape = rgb.shape[1:]
        rgb = rgb.reshape(3, -1)

        min_val = np.min(rgb, axis=0)
        max_val = np.max(rgb, axis=0)
        diff = max_val - min_val

        L = (max_val + min_val) / 2

        non_gray = diff > 0
        s = np.zeros_like(L)
        h = np.zeros_like(L)
        if non_gray.any():
            s[non_gray] = (diff[non_gray] / (2 - (max_val[non_gray] + min_val[non_gray])))
            dim_mask = np.logical_and(non_gray, L < .5)
            s[dim_mask] = (diff[dim_mask] / (max_val[dim_mask] + min_val[dim_mask]))

            h_ = self._rgb2hue(rgb[:, non_gray], diff[non_gray], max_val[non_gray])
            h[non_gray] = h_

        hsl = np.array([h, s, L])
        hsl = hsl.reshape((3,) + base_shape)
        if smol:
            hsl = hsl[:, 0]

        return hsl

    @classmethod
    def _hue2rgb(cls, v1, v2, h):

        h = np.array(h)
        v1 = np.array(v1)
        v2 = np.array(v2)
        h[h < 0] += 1
        h[h > 1] -= 1

        res = v1.copy()
        mask = 6*h < 1
        res[mask] = v1[mask] + (v2[mask] - v1[mask]) * 6*h[mask]
        rem = np.logical_not(mask)
        mask = np.logical_and(rem, 2*h < 1)
        res[mask] = v2[mask]
        rem = np.logical_and(rem, np.logical_not(mask))
        mask = np.logical_and(rem, 3*h < 2)
        res[mask] = v1[mask] + (v2[mask] - v1[mask]) * 6*(2/3 - h[mask])

        return res

    @classmethod
    def hsl_to_rgb(cls, h, s, l):

        hsl = np.array([h, s, l])
        smol = hsl.ndim == 1
        if smol:
            hsl = hsl[:, np.newaxis]
        base_shape = hsl.shape[1:]
        hsl = hsl.reshape((3,-1))

        non_gray = hsl[1] > 0
        dim = np.logical_and(non_gray, hsl[2] < .5)

        rgb = np.zeros_like(hsl)
        if non_gray.any():
            h, s, L = hsl
            v2 = np.zeros_like(h)
            v2[non_gray] = ((L + s) - (s * L))[non_gray]
            v2[dim] = (L * (1 + s))[dim]

            v1 = np.zeros_like(h)
            v1[non_gray] = 2 * L[non_gray] - v2[non_gray]

            for i,shift in enumerate([1/3, 0, -1/3]):
                rgb[i, non_gray] = cls._hue2rgb(v1[non_gray], v2[non_gray], h[non_gray] + shift)

        gray = np.logical_not(non_gray)
        for i in range(3):
            rgb[i, gray] = hsl[2, gray]

        rgb = rgb.reshape((3,) + base_shape)
        if smol:
            rgb = rgb[:, 0]

        return rgb * 255

    @classmethod
    def rgb_to_hsv(self, r, g, b):

        rgb = np.array([r, g, b]) / 255
        smol = rgb.ndim == 1
        if smol:
            rgb = rgb[:, np.newaxis]
        base_shape = rgb.shape[1:]
        rgb = rgb.reshape((3, -1))

        min_val = np.min(rgb, axis=0)
        max_val = np.max(rgb, axis=0)
        diff = max_val - min_val

        v = max_val

        non_gray = diff > 0
        s = np.zeros_like(v)
        h = np.zeros_like(v)
        if non_gray.any():
            s[non_gray] = (diff[non_gray] / max_val[non_gray])
            h_ = self._rgb2hue(rgb[:, non_gray], diff[non_gray], max_val[non_gray])
            h[non_gray] = h_

        hsl = np.array([h, s, v])
        hsl = hsl.reshape((3,) + base_shape)
        if smol:
            hsl = hsl[:, 0]

        return hsl

    @classmethod
    def hsv_to_hsl(cls, h, s, v):

        hsv = np.array([h, s, v])
        smol = hsv.ndim == 1
        if smol:
            hsv = hsv[:, np.newaxis]
        base_shape = hsv.shape[1:]
        hsv = hsv.reshape((3, -1))

        h, s, v = hsv
        max_val = np.array(v)
        diff = max_val * np.array(s)
        min_val = max_val - diff

        L = (max_val + min_val) / 2

        non_gray = diff > 0
        s = np.zeros_like(L)
        if non_gray.any():
            s[non_gray] = (diff[non_gray] / (2 - (max_val[non_gray] + min_val[non_gray])))
            dim_mask = np.logical_and(non_gray, L < .5)
            s[dim_mask] = (diff[dim_mask] / (max_val[dim_mask] + min_val[dim_mask]))

        hsl = np.array([h, s, L])
        hsl = hsl.reshape((3,) + base_shape)
        if smol:
            hsl = hsl[:, 0]

        return hsl

    @classmethod
    def hsv_to_rgb(cls, h, s, v):
        return cls.hsl_to_rgb(*cls.hsv_to_hsl(h, s, v))

    lab_scaling_reference = [95.0489, 100.0, 108.8840]
    @classmethod
    def xyz_to_lab(cls, x, y, z, scaling=None):

        xyz = np.array([x, y, z])
        smol = xyz.ndim == 1
        if smol:
            xyz = xyz[:, np.newaxis]
        base_shape = xyz.shape[1:]
        xyz = xyz.reshape((3, -1))

        if scaling is None:
            scaling = cls.lab_scaling_reference

        xyz /= np.array(scaling)[:, np.newaxis]

        mask = xyz > 0.008856
        xyz[mask] = xyz[mask] ** (1/3)
        not_max = np.logical_not(mask)
        xyz[not_max] = (7.787 * xyz[not_max]) + (16/116)

        L = 116 * xyz[1] - 16
        a = 500 * (xyz[0] - xyz[1])
        b = 200 * (xyz[1] - xyz[2])

        lab = np.array([L, a, b])
        lab = lab.reshape((3,) + base_shape)
        if smol:
            lab = lab[:, 0]

        return lab

    @classmethod
    def lab_to_xyz(cls, l, a, b, scaling=None):

        lab = np.array([l, a, b])
        smol = lab.ndim == 1
        if smol:
            lab = lab[:, np.newaxis]
        base_shape = lab.shape[1:]
        lab = lab.reshape((3, -1))

        if scaling is None:
            scaling = cls.lab_scaling_reference

        y = (lab[0] + 16) / 116
        x = lab[1] / 500 + y
        z = y - lab[2] / 200

        xyz = np.array([x, y, z])

        mask = xyz**3 > 0.008856
        xyz[mask] = xyz[mask] ** (3)
        not_max = np.logical_not(mask)
        xyz[not_max] = (xyz[not_max] - (16/116)) / 7.787

        xyz *= np.array(scaling)[:, np.newaxis]
        xyz = xyz.reshape((3,) + base_shape)
        if smol:
            xyz = xyz[:, 0]

        return xyz

    @classmethod
    def lab_to_lch(cls, l, a, b):
        c = np.linalg.norm([a, b], axis=0)
        h = np.arctan2(b, a)
        return np.array([l, c, h])
    @classmethod
    def lch_to_lab(cls, l, c, h):
        return np.array([
            l,
            np.cos(h) * c,
            np.sin(h) * c
        ])
    @classmethod
    def rgb_to_lab(cls, r, g, b, xyz_scaling=None):
        return cls.xyz_to_lab(*cls.rgb_to_xyz(r, g, b), scaling=xyz_scaling)
    @classmethod
    def lab_to_rgb(cls, l, a, b, xyz_scaling=None):
        return cls.xyz_to_rgb(*cls.lab_to_xyz(l, a, b, scaling=xyz_scaling))


"""
Basic layer for Schema validation that provides a superset of JSON schema validation
"""
import enum
import importlib
import re
import sys
import numbers

from . import core, Options as opts

__all__ = [
    "Schema"
]

class JSONSchemaTypes(enum.Enum):
    String = "string"
    Null = "null"
    Boolean = "boolean"
    Object = "object"
    Number = "number"
    Array = "array"

class TypeValidator:
    schema_types = JSONSchemaTypes

    def __init__(self, type_obj):
        self.types, self.validators = self.get_validators(type_obj)

    allow_imports = True
    base_type_map = {
        'str':str,
        'string':str,
        'int':int,
        'integer':int,
        'float':float,
        'real':float,
        'complex':complex,
        'bool':bool
    }
    @classmethod
    def resolve_typestr(cls, type_spec):
        if type_spec.lower() in cls.base_type_map:
            return cls.base_type_map[type_spec]
        else:
            type_spec = cls.base_type_map.get(type_spec, type_spec)
            if isinstance(type_spec, str):
                try:
                    type_spec = JSONSchemaTypes(type_spec)
                except ValueError:
                    mod_spec, type_spec = type_spec.rsplit('.')
                    if cls.allow_imports:
                        type_mod = importlib.import_module(mod_spec)
                    else:
                        if mod_spec in sys.modules:
                            type_mod = sys.modules[mod_spec]
                        else:
                            bits = mod_spec.split('.')
                            type_mod = sys.modules[bits[0]]
                            for t in bits[1:]:
                                type_mod = getattr(type_mod, t)
                    type_spec = getattr(type_mod, type_spec)
            return type_spec

    @classmethod
    def get_schema_type(cls, t):
        if t == cls.schema_types.Number:
            return (numbers.Number,)
        elif t == cls.schema_types.String:
            return (str,)
        elif t == cls.schema_types.Boolean:
            return (bool,)
        elif t == cls.schema_types.Null:
            return (type(None),)
        elif t == cls.schema_types.Object:
            return (core.is_dict_like,)
        elif t == cls.schema_types.Array:
            return (core.is_list_like,)
        else:
            raise ValueError(f"don't know what to do with {t}")

    @classmethod
    def prep_type_obj(cls, t):
        if isinstance(t, cls.schema_types):
            return cls.get_schema_type(t)
        else:
            return (t,)
    @classmethod
    def get_validators(cls, type_spec):
        if type_spec is None:
            return None, None

        if not core.is_list_like(type_spec):
            type_spec = [type_spec]

        type_spec = [
            cls.resolve_typestr(t)
                if isinstance(t, str) else
            t
            for t in type_spec
        ]

        type_spec = [
            tt
            for t in type_spec
            for tt in cls.prep_type_obj(t)
        ]

        types = tuple(
            t for t in type_spec
            if isinstance(t, type)
        )
        if len(types) == 0: types = None

        validators = tuple(
            t for t in type_spec
            if not isinstance(t, type) and t is not None
        )
        if len(validators) == 0: validators = None


        return types, validators

    def __repr__(self):
        cls = type(self)
        return f"{cls.__name__}(types={self.types}, validators={self.validators})"

    def validate(self, obj, throw=False):
        if self.types is not None and not isinstance(obj, self.types):
            return False
        if self.validators is not None and not all(
                v.validate(obj, throw=False)
                    if hasattr(v, 'validate') else
                v(obj)
                for v in self.validators
        ):
            return False
        return True

    def __call__(self, obj):
        return self.validate(obj)


class ValueValidator:
    __props__ = (
        "enum",
        "const",
        "multipleOf",
        "maximum",
        "minimum",
        "exclusiveMaximum",
        "exclusiveMinimum",
        "maxLength",
        "minLength",
        "pattern",
        "maxItems",
        "minItems",
        "uniqueItems",
        "maxProperties",
        "minProperties",
        "required",
        "validation_function"
    )
    def __init__(self,
                 enum=None,
                 const=None,
                 multipleOf=None,
                 maximum=None,
                 minimum=None,
                 exclusiveMaximum=None,
                 exclusiveMinimum=None,
                 maxLength=None,
                 minLength=None,
                 pattern=None,
                 maxItems=None,
                 minItems=None,
                 uniqueItems=None,
                 maxProperties=None,
                 minProperties=None,
                 required=None,
                 validation_function=None
                 ):
        self.tests = []
        self.enum = enum
        if self.enum is not None: self.tests.append(self._test_enum)
        self.const = const
        if self.const is not None: self.tests.append(self._test_const)
        self.multipleOf = multipleOf
        if self.multipleOf is not None: self.tests.append(self._test_multipleOf)
        self.maximum = maximum
        if self.maximum is not None: self.tests.append(self._test_maximum)
        self.minimum = minimum
        if self.minimum is not None: self.tests.append(self._test_minimum)
        self.exclusiveMaximum = exclusiveMaximum
        if self.exclusiveMaximum is not None: self.tests.append(self._test_exclusiveMaximum)
        self.exclusiveMinimum = exclusiveMinimum
        if self.exclusiveMinimum is not None: self.tests.append(self._test_exclusiveMinimum)
        self.maxLength = maxLength
        if self.maxLength is not None: self.tests.append(self._test_maxLength)
        self.minLength = minLength
        if self.minLength is not None: self.tests.append(self._test_minLength)
        self.pattern = pattern
        if self.pattern is not None: self.tests.append(self._test_pattern)
        self.maxItems = maxItems
        if self.maxItems is not None: self.tests.append(self._test_maxItems)
        self.minItems = minItems
        if self.minItems is not None: self.tests.append(self._test_minItems)
        self.uniqueItems = uniqueItems
        if self.uniqueItems is not None: self.tests.append(self._test_uniqueItems)
        self.validation_function = validation_function
        self.maxProperties = maxProperties
        if self.maxProperties is not None: self.tests.append(self._test_maxProperties)
        self.minProperties = minProperties
        if self.minProperties is not None: self.tests.append(self._test_minProperties)
        self.required = required
        if self.required is not None: self.tests.append(self._test_required)
        if self.validation_function is not None: self.tests.append(self._test_validation_function)

    def _test_validation_function(self, value): return not self.validation_function(value)
    def _test_enum(self, value): return value not in self.enum
    def _test_const(self, value): return value != self.const
    def _test_multipleOf(self, value): return (value % self.multipleOf) != 0
    def _test_maximum(self, value): return value > self.maximum
    def _test_minimum(self, value): return value < self.minimum
    def _test_exclusiveMaximum(self, value): return value >= self.exclusiveMaximum
    def _test_exclusiveMinimum(self, value): return value <= self.exclusiveMinimum
    def _test_maxLength(self, value): return len(value) > self.maxLength
    def _test_minLength(self, value): return len(value) < self.minLength
    def _test_maxItems(self, value): return len(value) > self.maxItems
    def _test_minItems(self, value): return len(value) < self.minItems
    def _test_maxProperties(self, value): return len(value.keys()) > self.maxProperties
    def _test_minProperties(self, value): return len(value.keys()) < self.minProperties
    def _test_required(self, value): return all(k in value for k in self.required)

    def _test_uniqueItems(self, value):
        for n, a in enumerate(value):
            for b in value[n + 1:]:
                if a == b: return True
        return False
    def _test_pattern(self, value):
        return not re.match(self.pattern, value)
    def validate(self, value, throw=False):
        for t in self.tests:
            if t(value): return False
        return True
    def __call__(self, obj):
        return self.validate(obj)


class Schema:
    """
    An object that represents a schema that can be used to test
    if an object matches that schema or not
    """
    type_validator = TypeValidator
    value_validator = ValueValidator

    def __init__(self, schema, optional_schema=None):
        self.schema = self.canonicalize_schema(schema, optional_schema)
        self._required = None

    @property
    def required_keys(self):
        if self._required is None:
            self._required = set(self.schema.get('required', []))
        return self._required
    # @property
    # def dependent_keys(self):
    #     return self.schema.get('dependentRequired', {})

    @classmethod
    def is_json_schema(self, schema):
        return "$schema" in schema

    @classmethod
    def _prep_schema_dict(cls, schema):
        if not core.is_dict_like(schema):
            if core.is_list_like(schema):
                schema = {
                    (
                        k
                        if len(k) != 2 or isinstance(k, str) else
                        k[0]
                    ):
                        (
                            None
                            if len(k) != 2 or isinstance(k, str) else
                            k[1]
                        )
                    for k in schema
                }
            else:
                raise ValueError(f"don't know how to canonicalize schema {schema}")
        return schema


    @classmethod
    def _prep_validators(cls, schema):
        new_schema = schema.copy()
        props = new_schema['properties'].copy()
        for k,v in props.items():
            if (not core.is_dict_like(v)) or 'type' not in v:
                v = {'type':v}
            else:
                v = v.copy()

            t = v['type']
            if core.is_dict_like(t):
                t = cls(t)

            v['type'] = cls.type_validator(t)
            if 'value' not in v:
                value_opts = opts.OptionsSet(v).filter(cls.value_validator)
                if len(value_opts) > 0:
                    v['value'] = cls.value_validator(**value_opts)

            props[k] = v
        new_schema['properties'] = props

        return new_schema


    json_schema_version = "https://json-schema.org/draft/2020-12/schema"
    @classmethod
    def canonicalize_schema(cls, schema, optional_schema=None):
        if schema is None:
            return None

        if not cls.is_json_schema(schema):
            schema = cls._prep_schema_dict(schema)

            old_schema = schema
            schema = {
                "$schema":cls.json_schema_version,
                "properties": {},
                "required": list(schema.keys())
            }
            props = schema["properties"]
            for k,v in old_schema.items():
                if v is None:
                    v = {"type":None}
                elif not core.is_dict_like(v):
                    v = {"type":v}
                props[k] = v

            if optional_schema is not None:
                optional_schema = cls._prep_schema_dict(optional_schema)
                for k,v in optional_schema.items():
                    if v is None:
                        v = {"type":None}
                    elif not core.is_dict_like(v):
                        v = {"type":v}
                    props[k] = v

        return cls._prep_validators(schema)
    @staticmethod
    def _get_prop(o, k):
        return (
            getattr(o, k)
                if not hasattr(o, '__getitem__') else
            o.__getitem__(k)
        )
    def _validate_entry(self, obj, k, v:dict, prop_getter=None, required=True, throw=False):
        match = True
        missing = False
        mistyped = False
        exc = None
        if prop_getter is None:
            prop_getter = self._get_prop
        try:
            t = prop_getter(obj, k)
        except (KeyError, AttributeError) as e:
            missing = True
            match = not required
            t = core.missing
            exc = e
        except (TypeError,) as e:
            mistyped = True
            match = False
            t = None
            exc = e
        else:
            if v is None:
                pass
            elif isinstance(v, (type, tuple)):
                match = isinstance(t, v)
            elif isinstance(v, Schema):
                match = v.validate(t, throw=throw)
            elif core.is_dict_like(v):
                validator = v['type']
                if validator is not None:
                    match = v['type'](t)
                    if match and 'value' in v:
                        match = v['value'](t)
            else:
                match = v(t)

        if not match and throw:
            if missing:
                raise KeyError("object {} doesn't match schema {}; key {} is missing".format(
                    obj, self, k
                )) from exc
            elif mistyped:
                raise KeyError("object {} doesn't match schema {}; doesn't support attributes".format(
                    obj, self
                )) from exc
            else:
                raise KeyError("object {} doesn't match schema {}; value {} doesn't match schema type {}".format(
                    obj, self, t, v
                )) from exc
        return (t, match)

    def validate(self, obj, throw=True):
        """
        Validates that `obj` matches the provided schema
        and throws an error if not

        :param obj:
        :type obj:
        :param throw:
        :type throw:
        :return:
        :rtype:
        """

        for k,v in self.schema['properties'].items():
            if not self._validate_entry(obj, k, v, required=k in self.required_keys, throw=throw)[1]:
                return False

        return True

    def to_dict(self, obj, throw=True, ignore_invalid=False):
        """
        Converts `obj` into a plain `dict` representation

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        res = {}
        for k, v in self.schema['properties'].items():
            t, m = self._validate_entry(obj, k, v, required=k in self.required_keys, throw=throw)
            if not m and not ignore_invalid:
                return None
            elif t is not core.missing:
                res[k] = t
        # if self.sub_schema is not None:
        #     for k, v in self.sub_schema.items():
        #         t, m = self._validate_entry(obj, k, v, prop_getter, throw=False)
        #         if m:
        #             res[k] = t
        return res

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.schema)


"""
The Coordinerds package implements stuff for dealing with coordinates and generalized coordinate systems

It provides a semi-symbolic way to represent a CoordinateSystem and a CoordinateSet that provides coordinates within a
coordinate system. An extensible system for converting between coordinate systems and is provided.

The basic design of the package is set up so that one creates a `CoordinateSet` object, which in turn tracks its `CoordinateSystem`.
A `CoordinateSet` is a subclass of `np.ndarray`, and so any operation that works for a `np.ndarray` will work in turn for `CoordinateSet`.
This provides a large amount flexibility.

The `CoordinateSystem` object handles much of the heavy lifting for a `CoordinateSet`.
Conversions between different systems are implemented by a `CoordinateSystemConverter`.
Chained conversions are not _currently_ supported, but might well become supported in the future.
"""


__all__ = []
from .CoordinateSystems import *; from .CoordinateSystems import __all__ as exposed
__all__ += exposed
# from .CoordinateTransformations import __all__ as exposed
# __all__ += exposed
from .Conveniences import *; from .Conveniences import __all__ as exposed
__all__ += exposed
from .Internals import *; from .Internals import __all__ as exposed
__all__ += exposed
from .ZMatrices import *; from .ZMatrices import __all__ as exposed
__all__ += exposed
from .Labels import *; from .Labels import __all__ as exposed
__all__ += exposed
from .Generators import *; from .Generators import __all__ as exposed
__all__ += exposed
from .Pruning import *; from .Pruning import __all__ as exposed
__all__ += exposed
from .Redundant import *; from .Redundant import __all__ as exposed
__all__ += exposed


"""
Provides utilities for working with objects, defaults, etc. to abstract away common idoms
"""

__all__ = []
from .core import *; from .core import __all__ as _all
__all__ += _all
from .FileHelpers import *; from .FileHelpers import __all__ as _all
__all__ += _all
from .Loggers import *; from .Loggers import __all__ as _all
__all__ += _all
from .Options import *; from .Options import __all__ as _all
__all__ += _all
from .Redirects import *; from .Redirects import __all__ as _all
__all__ += _all
from .Schema import *; from .Schema import __all__ as _all
__all__ += _all


"""Provides a little timer for testing

"""
import timeit, functools, time, sys, inspect
import uuid
from collections import deque

__all__ = ["Timer"]

class Timer:

    tag_printing_times = {}
    def __init__(self, tag=None, file=sys.stderr, rounding=6, message=None, format=None, print_times=-1, number=None, globals=None, **kw):
        self.kw = kw
        self.number = number
        self.file = file
        self.tag = tag
        self.message = message
        self.format = format
        self.rounding = rounding
        self.checkpoints = deque()
        self.print_times = print_times
        self.latest = None
        self.globals = globals
        self.laps = []

    def get_time_list(self, time_elapsed):
        run_time = [0, 0, time_elapsed]
        if run_time[2] > 60:
            run_time[1] = int(run_time[2] / 60)
            run_time[2] = run_time[2] % 60
            if run_time[1] > 60:
                run_time[0] = int(run_time[1] / 60)
                run_time[1] = run_time[1] % 60
        return run_time

    def start(self):
        self.checkpoints.append(time.time())

    def stop(self):
        t = time.time()
        cp = self.checkpoints.pop()
        return t - cp

    def log(self):
        t = time.time()
        self.laps.append([self.checkpoints[-1], t])

    def __enter__(self):
        self.start()
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.latest = self.stop()
        self.print_timing(self.latest)

    default_time_format = "{hours}:{minutes:0>2}:{seconds:0>{width}.{rounding}f}"
    def format_time(self, timelist, format=None, rounding=6):
        if format is None:
            format = self.format
        if format is None:
            format = self.default_time_format
        if rounding is not None:
            rounding = self.rounding
        if rounding is None:
            rounding = len(str(timelist[-1])) - len(str(int(timelist[-1]))) + 1
        return format.format(hours=timelist[0], minutes=timelist[1], seconds=timelist[2],
                          width=rounding+3,
                          rounding=rounding
                          )


    default_message = "{tag}: took {avg} per loop with {tot} overall"
    def format_timing(self, time_elapsed, *, message=None, format=None, rounding=None, tag=None, steps=None):
        run_time = self.get_time_list(time_elapsed)
        if tag is None:
            tag = self.tag
        if steps is None:
            steps = self.number
        if steps is None:
            steps = 1
        run_time_averaged = self.get_time_list(time_elapsed / steps)
        if message is None:
            message = self.message
        if message is None:
            message = self.default_message
        if rounding is None:
            rounding = self.rounding
        run_time_averaged = self.format_time(run_time_averaged, format=format, rounding=rounding)
        run_time = self.format_time(run_time, format=format, rounding=rounding)
        return message.format(
            tag=tag,
            avg=run_time_averaged,
            tot=run_time
        )

    def print_timing(self, time_elapsed, tag=None, steps=None):
        if tag is None:
            tag = self.tag
        if tag not in self.tag_printing_times:
            self.tag_printing_times[tag] = 0
        if self.print_times < 0 or self.tag_printing_times[tag] < self.print_times:
            print(self.format_timing(time_elapsed, tag=tag, steps=steps), file=self.file)
            self.tag_printing_times[tag] += 1

    def timeit(self, stmt, *args, **kwargs):
        globals = self.globals
        val = None
        number = self.number
        tag = self.tag
        id = "_" + str(uuid.uuid4()).replace("-", "_")
        if isinstance(stmt, str):
            if globals is None:
                globals = inspect.stack(1)[0].frame.f_locals
            timer = timeit.Timer(stmt)
            if tag is None:
                if hasattr(stmt, '__name__'):
                    tag = stmt.__name__
                else:
                    tag = str(stmt)
        else:
            if globals is None:
                try:
                    globals = stmt.__globals__
                except AttributeError:
                    globals = inspect.stack(1)[0].frame.f_locals

            globals[f'_{id}_func'] = stmt
            globals[f'_{id}_args'] = args
            globals[f'_{id}_kwargs'] = kwargs
            timer = timeit.Timer(
                f"""
global _{id}_val
_{id}_val = _{id}_func(*_{id}_args, **_{id}_kwargs)
""", globals=globals,
                **self.kw
            )
            if tag is None:
                if hasattr(stmt, '__name__'):
                    tag = stmt.__name__
                else:
                    tag = str(stmt)

        if number is None:
            number, elapsed_time = timer.autorange()
        else:
            elapsed_time = timer.timeit(number)

        if not isinstance(stmt, str):
            val = globals[f"_{id}_val"]

        self.print_timing(elapsed_time, steps=number, tag=tag)
        return val, elapsed_time, number

    def __call__(self, fn): # for use as a decorator
        @functools.wraps(fn)
        def timed_fn(*args, **kwargs):
            return self.timeit(fn, *args, **kwargs)[0]
        return timed_fn


"""
A place to store all utilities related to combinatorics.
Currently only contains a subpackage for working with permutations.
That package is used in the `BasisReps` work in `Psience`.

It handles both integer partitions and unique permutations.
It

Might be worth extending to handle more lattice-path stuff.

:examples:
Get all partitions of `3`

```python
IntegerPartitioner.partitions(3)
```

Count the number of partitions up to 10

```python
IntegerPartitioner.count_partitions(10)
```

Count the number of partitions of `4` of length exactly `2`
with minimum component between `2` and `4`

```python
IntegerPartitioner.count_exact_length_partitions_in_range(4, 4, 2, 2)
```

Get all unique permutations of the first integer partition of 10

```python
lens, parts = IntegerPartitioner.partitions(10, pad=True, return_lens=True)
perms = UniquePermutations(parts[0]).permutations()
```

Get all 9D state vectors and indices for states with up to 4 quanta of excitation

```python
SymmetricGroupGenerator(9).get_terms([1, 2, 3, 4])
```
"""

__all__ = []
from .Permutations import *; from .Permutations import __all__ as exposed
__all__ += exposed
from .Sequences import *; from .Sequences import __all__ as exposed
__all__ += exposed
from .YoungTableaux import *; from .YoungTableaux import __all__ as exposed
__all__ += exposed


"""
Defines a set of formatting utilities
"""

__all__ = []
from .TemplateWriter import *; from .TemplateWriter import __all__ as exposed
__all__ += exposed
from .TemplateEngine import *; from .TemplateEngine import __all__ as exposed
__all__ += exposed
from .TeXWriter import *; from .TeXWriter import __all__ as exposed
__all__ += exposed
from .FileMatcher import *; from .FileMatcher import __all__ as exposed
__all__ += exposed
from .TableFormatters import *; from .TableFormatters import __all__ as exposed
__all__ += exposed
from .Conveniences import *; from .Conveniences import __all__ as exposed
__all__ += exposed